import os
from pathlib import Path

from slither import Slither
from slither.core.expressions import Literal
from slither.utils.upgradeability import (
    compare,
    get_proxy_implementation_var,
    get_proxy_implementation_slot,
)

SLITHER_ROOT = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data" / "upgradeability_util"


def test_upgrades_compare(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.2")

    sl = Slither(os.path.join(TEST_DATA_DIR, "TestUpgrades-0.8.2.sol"), solc=solc_path)
    v1 = sl.get_contract_from_name("ContractV1")[0]
    v2 = sl.get_contract_from_name("ContractV2")[0]
    (
        missing_vars,
        new_vars,
        tainted_vars,
        new_funcs,
        modified_funcs,
        tainted_funcs,
        tainted_contracts,
    ) = compare(v1, v2, include_external=True)
    assert len(missing_vars) == 0
    assert new_vars == [v2.get_state_variable_from_name("stateC")]
    assert tainted_vars == [
        v2.get_state_variable_from_name("bug"),
    ]
    assert new_funcs == [
        v2.get_function_from_signature("i()"),
        v2.get_function_from_signature("erc20Transfer(address,address,uint256)"),
    ]
    assert modified_funcs == [v2.get_function_from_signature("checkB()")]
    assert tainted_funcs == [
        v2.get_function_from_signature("h()"),
    ]
    erc20 = sl.get_contract_from_name("ERC20")[0]
    assert len(tainted_contracts) == 1
    assert tainted_contracts[0].contract == erc20
    assert set(tainted_contracts[0].tainted_functions) == {
        erc20.get_function_from_signature("transfer(address,uint256)"),
        erc20.get_function_from_signature("_transfer(address,address,uint256)"),
        erc20.get_function_from_signature("_burn(address,uint256)"),
        erc20.get_function_from_signature("balanceOf(address)"),
        erc20.get_function_from_signature("_mint(address,uint256)"),
    }
    assert tainted_contracts[0].tainted_variables == [
        erc20.get_state_variable_from_name("_balances")
    ]


def test_upgrades_implementation_var(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.2")
    sl = Slither(os.path.join(TEST_DATA_DIR, "TestUpgrades-0.8.2.sol"), solc=solc_path)

    erc_1967_proxy = sl.get_contract_from_name("ERC1967Proxy")[0]
    storage_proxy = sl.get_contract_from_name("InheritedStorageProxy")[0]
    upgradeable_proxy = sl.get_contract_from_name("UpgradeableProxy")[0]
    transparent_proxy = sl.get_contract_from_name("TransparentUpgradeableProxy")[0]

    target = get_proxy_implementation_var(erc_1967_proxy)
    slot = get_proxy_implementation_slot(erc_1967_proxy)
    assert target == erc_1967_proxy.get_state_variable_from_name("_IMPLEMENTATION_SLOT")
    assert slot.slot == 0x360894A13BA1A3210667C828492DB98DCA3E2076CC3735A920A3CA505D382BBC
    target = get_proxy_implementation_var(storage_proxy)
    slot = get_proxy_implementation_slot(storage_proxy)
    assert target == storage_proxy.get_state_variable_from_name("implementation")
    assert slot.slot == 1
    target = get_proxy_implementation_var(transparent_proxy)
    slot = get_proxy_implementation_slot(transparent_proxy)
    assert target == upgradeable_proxy.get_state_variable_from_name("_IMPLEMENTATION_SLOT")
    assert slot.slot == 0x360894A13BA1A3210667C828492DB98DCA3E2076CC3735A920A3CA505D382BBC

    solc_path = solc_binary_path("0.5.0")
    sl = Slither(os.path.join(TEST_DATA_DIR, "TestUpgrades-0.5.0.sol"), solc=solc_path)

    eip_1822_proxy = sl.get_contract_from_name("EIP1822Proxy")[0]
    # zos_proxy = sl.get_contract_from_name("ZosProxy")[0]
    master_copy_proxy = sl.get_contract_from_name("MasterCopyProxy")[0]
    synth_proxy = sl.get_contract_from_name("SynthProxy")[0]

    target = get_proxy_implementation_var(eip_1822_proxy)
    slot = get_proxy_implementation_slot(eip_1822_proxy)
    assert target not in eip_1822_proxy.state_variables_ordered
    assert target.name == "contractLogic" and isinstance(target.expression, Literal)
    assert (
        target.expression.value
        == "0xc5f16f0fcc639fa48a6947836d9850f504798523bf8c9a3a87d5876cf622bcf7"
    )
    assert slot.slot == 0xC5F16F0FCC639FA48A6947836D9850F504798523BF8C9A3A87D5876CF622BCF7
    # # The util fails with this proxy due to how Slither parses assembly w/ Solidity versions < 0.6.0 (see issue #1775)
    # target = get_proxy_implementation_var(zos_proxy)
    # slot = get_proxy_implementation_slot(zos_proxy)
    # assert target == zos_proxy.get_state_variable_from_name("IMPLEMENTATION_SLOT")
    # assert slot.slot == 0x7050C9E0F4CA769C69BD3A8EF740BC37934F8E2C036E5A723FD8EE048ED3F8C3
    target = get_proxy_implementation_var(master_copy_proxy)
    slot = get_proxy_implementation_slot(master_copy_proxy)
    assert target == master_copy_proxy.get_state_variable_from_name("masterCopy")
    assert slot.slot == 0
    target = get_proxy_implementation_var(synth_proxy)
    slot = get_proxy_implementation_slot(synth_proxy)
    assert target == synth_proxy.get_state_variable_from_name("target")
    assert slot.slot == 1



"""
Main module
"""

import json
import logging
import os
import pathlib
import posixpath
import re
from collections import defaultdict
from typing import TypeVar

from crytic_compile import CryticCompile
from crytic_compile.utils.naming import Filename

from slither.core.declarations.contract_level import ContractLevel
from slither.core.compilation_unit import SlitherCompilationUnit
from slither.core.context.context import Context
from slither.core.declarations import Contract, FunctionContract
from slither.core.declarations.top_level import TopLevel
from slither.core.source_mapping.source_mapping import SourceMapping, Source
from slither.slithir.variables import Constant
from slither.utils.colors import red
from slither.utils.sarif import read_triage_info
from slither.utils.source_mapping import get_definition, get_references, get_all_implementations

logger = logging.getLogger("Slither")
logging.basicConfig()


def _relative_path_format(path: str) -> str:
    """
    Strip relative paths of "." and ".."
    """
    return path.split("..")[-1].strip(".").strip("/")


class SlitherCore(Context):
    """
    Slither static analyzer
    """

    def __init__(self) -> None:
        super().__init__()

        self._filename: str | None = None
        self._raw_source_code: dict[str, str] = {}
        self._source_code_to_line: dict[str, list[str]] | None = None

        self._previous_results_filename: str = "slither.db.json"

        # TODO: add cli flag to set these variables
        self.sarif_input: str = "export.sarif"
        self.sarif_triage: str = "export.sarif.sarifexplorer"
        self._results_to_hide: list = []
        self._previous_results: list = []
        # From triaged result
        self._previous_results_ids: set[str] = set()
        # Every slither object has a list of result from detector
        # Because of the multiple compilation support, we might analyze
        # Multiple time the same result, so we remove duplicates
        self._currently_seen_resuts: set[str] = set()
        self._paths_to_filter: set[str] = set()
        self._paths_to_include: set[str] = set()

        self._crytic_compile: CryticCompile | None = None

        self._generate_patches = False
        self._exclude_dependencies = False

        self._markdown_root = ""

        # If set to true, slither will not catch errors during parsing
        self._disallow_partial: bool = False
        self._skip_assembly: bool = False

        self._show_ignored_findings = False

        # Maps from file to detector name to the start/end ranges for that detector.
        # Infinity is used to signal a detector has no end range.
        self._ignore_ranges: dict[str, dict[str, list[tuple[int, ...]]]] = defaultdict(
            lambda: defaultdict(lambda: [(-1, -1)])
        )

        self._compilation_units: list[SlitherCompilationUnit] = []

        self._contracts: list[Contract] = []
        self._contracts_derived: list[Contract] = []

        self._offset_to_min_offset: dict[Filename, dict[int, set[int]]] | None = None
        self._offset_to_objects: dict[Filename, dict[int, set[SourceMapping]]] | None = None
        self._offset_to_references: dict[Filename, dict[int, set[Source]]] | None = None
        self._offset_to_implementations: dict[Filename, dict[int, set[Source]]] | None = None
        self._offset_to_definitions: dict[Filename, dict[int, set[Source]]] | None = None

        # Line prefix is used during the source mapping generation
        # By default we generate file.sol#1
        # But we allow to alter this (ex: file.sol:1) for vscode integration
        self.line_prefix: str = "#"

        # Use by the echidna printer
        # If true, partial analysis is allowed
        self.no_fail = False

        self.skip_data_dependency = False

    @property
    def compilation_units(self) -> list[SlitherCompilationUnit]:
        return list(self._compilation_units)

    def add_compilation_unit(self, compilation_unit: SlitherCompilationUnit):
        self._compilation_units.append(compilation_unit)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Contracts
    ###################################################################################
    ###################################################################################

    @property
    def contracts(self) -> list[Contract]:
        if not self._contracts:
            all_contracts = [
                compilation_unit.contracts for compilation_unit in self._compilation_units
            ]
            self._contracts = [item for sublist in all_contracts for item in sublist]
        return self._contracts

    @property
    def contracts_derived(self) -> list[Contract]:
        if not self._contracts_derived:
            all_contracts = [
                compilation_unit.contracts_derived for compilation_unit in self._compilation_units
            ]
            self._contracts_derived = [item for sublist in all_contracts for item in sublist]
        return self._contracts_derived

    def get_contract_from_name(self, contract_name: str | Constant) -> list[Contract]:
        """
            Return a contract from a name
        Args:
            contract_name (str): name of the contract
        Returns:
            Contract
        """
        contracts = []
        for compilation_unit in self._compilation_units:
            contracts += compilation_unit.get_contract_from_name(contract_name)
        return contracts

    ###################################################################################
    ###################################################################################
    # region Source code
    ###################################################################################
    ###################################################################################

    @property
    def source_code(self) -> dict[str, str]:
        """{filename: source_code (str)}: source code"""
        return self._raw_source_code

    @property
    def filename(self) -> str | None:
        """str: Filename."""
        return self._filename

    @filename.setter
    def filename(self, filename: str):
        self._filename = filename

    def add_source_code(self, path: str) -> None:
        """
        :param path:
        :return:
        """
        if self.crytic_compile and path in self.crytic_compile.src_content:
            self.source_code[path] = self.crytic_compile.src_content[path]
        else:
            with open(path, encoding="utf8", newline="") as f:
                self.source_code[path] = f.read()

        self.parse_ignore_comments(path)

    @property
    def markdown_root(self) -> str:
        return self._markdown_root

    def print_functions(self, d: str):
        """
        Export all the functions to dot files
        """
        for compilation_unit in self._compilation_units:
            for c in compilation_unit.contracts:
                for f in c.functions:
                    f.cfg_to_dot(os.path.join(d, f"{c.name}.{f.name}.dot"))

    def _compute_offsets_from_thing(self, thing: SourceMapping):
        definition = get_definition(thing, self.crytic_compile)
        references = get_references(thing)
        implementations = get_all_implementations(thing, self.contracts)

        # Create the offset mapping
        for offset in range(definition.start, definition.end + 1):
            self._offset_to_min_offset[definition.filename][offset].add(definition.start)

        is_declared_function = (
            isinstance(thing, FunctionContract) and thing.contract_declarer == thing.contract
        )

        should_add_to_objects = (
            isinstance(thing, (TopLevel, Contract))
            or is_declared_function
            or (isinstance(thing, ContractLevel) and not isinstance(thing, FunctionContract))
        )

        if should_add_to_objects:
            self._offset_to_objects[definition.filename][definition.start].add(thing)

        self._offset_to_definitions[definition.filename][definition.start].add(definition)
        self._offset_to_implementations[definition.filename][definition.start].update(
            implementations
        )
        self._offset_to_references[definition.filename][definition.start] |= set(references)

        # For references
        should_add_to_objects = (
            isinstance(thing, TopLevel)
            or is_declared_function
            or (isinstance(thing, ContractLevel) and not isinstance(thing, FunctionContract))
        )

        for ref in references:
            for offset in range(ref.start, ref.end + 1):
                self._offset_to_min_offset[definition.filename][offset].add(ref.start)

            if should_add_to_objects:
                self._offset_to_objects[definition.filename][ref.start].add(thing)

            if is_declared_function:
                # Only show the nearest lexical definition for declared contract-level functions
                if (
                    thing.contract.source_mapping.start
                    < ref.start
                    < thing.contract.source_mapping.end
                ):
                    self._offset_to_definitions[ref.filename][ref.start].add(definition)

            else:
                self._offset_to_definitions[ref.filename][ref.start].add(definition)

            self._offset_to_implementations[ref.filename][ref.start].update(implementations)
            self._offset_to_references[ref.filename][ref.start] |= set(references)

    def _compute_offsets_to_ref_impl_decl(self):
        self._offset_to_references = defaultdict(lambda: defaultdict(lambda: set()))
        self._offset_to_definitions = defaultdict(lambda: defaultdict(lambda: set()))
        self._offset_to_implementations = defaultdict(lambda: defaultdict(lambda: set()))
        self._offset_to_objects = defaultdict(lambda: defaultdict(lambda: set()))
        self._offset_to_min_offset = defaultdict(lambda: defaultdict(lambda: set()))

        for compilation_unit in self._compilation_units:
            for contract in compilation_unit.contracts:
                self._compute_offsets_from_thing(contract)

                for function in contract.functions_declared:
                    self._compute_offsets_from_thing(function)
                    for variable in function.local_variables:
                        self._compute_offsets_from_thing(variable)
                for modifier in contract.modifiers_declared:
                    self._compute_offsets_from_thing(modifier)
                    for variable in modifier.local_variables:
                        self._compute_offsets_from_thing(variable)

                for var in contract.state_variables:
                    self._compute_offsets_from_thing(var)

                for st in contract.structures:
                    self._compute_offsets_from_thing(st)

                for enum in contract.enums:
                    self._compute_offsets_from_thing(enum)

                for event in contract.events:
                    self._compute_offsets_from_thing(event)

                for typ in contract.type_aliases:
                    self._compute_offsets_from_thing(typ)

            for enum in compilation_unit.enums_top_level:
                self._compute_offsets_from_thing(enum)
            for event in compilation_unit.events_top_level:
                self._compute_offsets_from_thing(event)
            for function in compilation_unit.functions_top_level:
                self._compute_offsets_from_thing(function)
            for st in compilation_unit.structures_top_level:
                self._compute_offsets_from_thing(st)
            for var in compilation_unit.variables_top_level:
                self._compute_offsets_from_thing(var)
            for typ in compilation_unit.type_aliases.values():
                self._compute_offsets_from_thing(typ)
            for err in compilation_unit.custom_errors:
                self._compute_offsets_from_thing(err)
            for event in compilation_unit.events_top_level:
                self._compute_offsets_from_thing(event)
            for import_directive in compilation_unit.import_directives:
                self._compute_offsets_from_thing(import_directive)
            for pragma in compilation_unit.pragma_directives:
                self._compute_offsets_from_thing(pragma)

    T = TypeVar("T", Source, SourceMapping)

    def _get_offset(
        self, mapping: dict[Filename, dict[int, set[T]]], filename_str: str, offset: int
    ) -> set[T]:
        """Get the Source/SourceMapping referenced by the offset.

        For performance reasons, references are only stored once at the lowest offset.
        It uses the _offset_to_min_offset mapping to retrieve the correct offsets.
        As multiple definitions can be related to the same offset, we retrieve all of them.

        :param mapping: Mapping to search for (objects. references, ...)
        :param filename_str: Filename to consider
        :param offset: Look-up offset
        :raises IndexError: When the start offset is not found
        :return: The corresponding set of Source/SourceMapping
        """
        filename: Filename = self.crytic_compile.filename_lookup(filename_str)

        start_offsets = self._offset_to_min_offset[filename][offset]
        if not start_offsets:
            msg = f"Unable to find reference for offset {offset}"
            raise IndexError(msg)

        results = set()
        for start_offset in start_offsets:
            results |= mapping[filename][start_offset]

        return results

    def offset_to_references(self, filename_str: str, offset: int) -> set[Source]:
        if self._offset_to_references is None:
            self._compute_offsets_to_ref_impl_decl()

        return self._get_offset(self._offset_to_references, filename_str, offset)

    def offset_to_implementations(self, filename_str: str, offset: int) -> set[Source]:
        if self._offset_to_implementations is None:
            self._compute_offsets_to_ref_impl_decl()

        return self._get_offset(self._offset_to_implementations, filename_str, offset)

    def offset_to_definitions(self, filename_str: str, offset: int) -> set[Source]:
        if self._offset_to_definitions is None:
            self._compute_offsets_to_ref_impl_decl()

        return self._get_offset(self._offset_to_definitions, filename_str, offset)

    def offset_to_objects(self, filename_str: str, offset: int) -> set[SourceMapping]:
        if self._offset_to_objects is None:
            self._compute_offsets_to_ref_impl_decl()

        return self._get_offset(self._offset_to_objects, filename_str, offset)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Filtering results
    ###################################################################################
    ###################################################################################

    def parse_ignore_comments(self, file: str) -> None:
        # The first time we check a file, find all start/end ignore comments and memoize them.
        line_number = 1
        while True:
            line_text = self.crytic_compile.get_code_from_line(file, line_number)
            if line_text is None:
                break

            start_regex = r"^\s*//\s*slither-disable-start\s*([a-zA-Z0-9_,-]*)"
            end_regex = r"^\s*//\s*slither-disable-end\s*([a-zA-Z0-9_,-]*)"
            start_match = re.findall(start_regex, line_text.decode("utf8"))
            end_match = re.findall(end_regex, line_text.decode("utf8"))

            if start_match:
                ignored = start_match[0].split(",")
                if ignored:
                    for check in ignored:
                        vals = self._ignore_ranges[file][check]
                        if len(vals) == 0 or vals[-1][1] != float("inf"):
                            # First item in the array, or the prior item is fully populated.
                            self._ignore_ranges[file][check].append((line_number, float("inf")))
                        else:
                            logger.error(
                                f"Consecutive slither-disable-starts without slither-disable-end in {file}#{line_number}"
                            )
                            return

            if end_match:
                ignored = end_match[0].split(",")
                if ignored:
                    for check in ignored:
                        vals = self._ignore_ranges[file][check]
                        if len(vals) == 0 or vals[-1][1] != float("inf"):
                            logger.error(
                                f"slither-disable-end without slither-disable-start in {file}#{line_number}"
                            )
                            return
                        self._ignore_ranges[file][check][-1] = (vals[-1][0], line_number)

            line_number += 1

    def has_ignore_comment(self, r: dict) -> bool:
        """
        Check if the result has an ignore comment in the file or on the preceding line, in which
        case, it is not valid
        """
        if not self.crytic_compile:
            return False
        mapping_elements_with_lines = (
            (
                posixpath.normpath(elem["source_mapping"]["filename_absolute"]),
                elem["source_mapping"]["lines"],
            )
            for elem in r["elements"]
            if "source_mapping" in elem
            and "filename_absolute" in elem["source_mapping"]
            and "lines" in elem["source_mapping"]
            and len(elem["source_mapping"]["lines"]) > 0
        )

        for file, lines in mapping_elements_with_lines:
            # Check if result is within an ignored range.
            ignore_ranges = self._ignore_ranges[file][r["check"]] + self._ignore_ranges[file]["all"]
            for start, end in ignore_ranges:
                # The full check must be within the ignore range to be ignored.
                if start < lines[0] and end > lines[-1]:
                    return True

            # Check for next-line matchers.
            ignore_line_index = min(lines) - 1
            ignore_line_text = self.crytic_compile.get_code_from_line(file, ignore_line_index)
            if ignore_line_text:
                match = re.findall(
                    r"^\s*//\s*slither-disable-next-line\s*([a-zA-Z0-9_,-]*)",
                    ignore_line_text.decode("utf8"),
                )
                if match:
                    ignored = match[0].split(",")
                    if ignored and ("all" in ignored or any(r["check"] == c for c in ignored)):
                        return True

        return False

    def valid_result(self, r: dict) -> bool:
        """
        Check if the result is valid
        A result is invalid if:
            - All its source paths belong to the source path filtered
            - Or a similar result was reported and saved during a previous run
            - The --exclude-dependencies flag is set and results are only related to dependencies
            - There is an ignore comment on the preceding line or in the file
        """

        # Remove duplicate due to the multiple compilation support
        if r["id"] in self._currently_seen_resuts:
            return False
        self._currently_seen_resuts.add(r["id"])

        source_mapping_elements = [
            elem["source_mapping"].get("filename_absolute", "unknown")
            for elem in r["elements"]
            if "source_mapping" in elem
        ]

        # Use POSIX-style paths so that filter_paths|include_paths works across different
        # OSes. Convert to a list so elements don't get consumed and are lost
        # while evaluating the first pattern
        source_mapping_elements = [
            pathlib.Path(x).resolve().as_posix() if x else x for x in source_mapping_elements
        ]
        (matching, paths, msg_err) = (
            (True, self._paths_to_include, "--include-paths")
            if self._paths_to_include
            else (False, self._paths_to_filter, "--filter-paths")
        )

        for path in paths:
            try:
                if any(
                    bool(re.search(_relative_path_format(path), src_mapping))
                    for src_mapping in source_mapping_elements
                ):
                    matching = not matching
                    break
            except re.error:
                logger.error(
                    f"Incorrect regular expression for {msg_err} {path}."
                    "\nSlither supports the Python re format"
                    ": https://docs.python.org/3/library/re.html"
                )

        if r["elements"] and matching:
            return False

        if self._show_ignored_findings:
            return True
        if self.has_ignore_comment(r):
            return False
        if r["id"] in self._previous_results_ids:
            return False
        if r["elements"] and self._exclude_dependencies:
            if all(element["source_mapping"]["is_dependency"] for element in r["elements"]):
                return False
        # Conserve previous result filtering. This is conserved for compatibility, but is meant to be removed
        if r["description"] in [pr["description"] for pr in self._previous_results]:
            return False

        return True

    def load_previous_results(self) -> None:
        self.load_previous_results_from_sarif()

        filename = self._previous_results_filename
        try:
            if os.path.isfile(filename):
                with open(filename, encoding="utf8") as f:
                    self._previous_results = json.load(f)
                    if self._previous_results:
                        for r in self._previous_results:
                            if "id" in r:
                                self._previous_results_ids.add(r["id"])

        except json.decoder.JSONDecodeError:
            logger.error(red(f"Impossible to decode {filename}. Consider removing the file"))

    def load_previous_results_from_sarif(self) -> None:
        sarif = pathlib.Path(self.sarif_input)
        triage = pathlib.Path(self.sarif_triage)

        if not sarif.exists():
            return
        if not triage.exists():
            return

        triaged = read_triage_info(sarif, triage)

        for id_triaged in triaged:
            self._previous_results_ids.add(id_triaged)

    def write_results_to_hide(self) -> None:
        if not self._results_to_hide:
            return
        filename = self._previous_results_filename
        with open(filename, "w", encoding="utf8") as f:
            results = self._results_to_hide + self._previous_results
            json.dump(results, f)

    def save_results_to_hide(self, results: list[dict]) -> None:
        self._results_to_hide += results
        self.write_results_to_hide()

    def add_path_to_filter(self, path: str):
        """
        Add path to filter
        Path are used through direct comparison (no regex)
        """
        self._paths_to_filter.add(path)

    def add_path_to_include(self, path: str):
        """
        Add path to include
        Path are used through direct comparison (no regex)
        """
        self._paths_to_include.add(path)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Crytic compile
    ###################################################################################
    ###################################################################################

    @property
    def crytic_compile(self) -> CryticCompile:
        return self._crytic_compile  # type: ignore

    # endregion
    ###################################################################################
    ###################################################################################
    # region Format
    ###################################################################################
    ###################################################################################

    @property
    def generate_patches(self) -> bool:
        return self._generate_patches

    @generate_patches.setter
    def generate_patches(self, p: bool):
        self._generate_patches = p

    # endregion
    ###################################################################################
    ###################################################################################
    # region Internals
    ###################################################################################
    ###################################################################################

    @property
    def disallow_partial(self) -> bool:
        """
        Return true if partial analyses are disallowed
        For example, codebase with duplicate names will lead to partial analyses

        :return:
        """
        return self._disallow_partial

    @property
    def skip_assembly(self) -> bool:
        return self._skip_assembly

    @property
    def show_ignore_findings(self) -> bool:
        return self._show_ignored_findings

    # endregion



from slither_my_plugin.detectors.example import Example

from slither.detectors.abstract_detector import AbstractDetector
from slither.printers.abstract_printer import AbstractPrinter


def make_plugin() -> tuple[list[type[AbstractDetector]], list[type[AbstractPrinter]]]:
    plugin_detectors = [Example]
    plugin_printers: list[type[AbstractPrinter]] = []

    return plugin_detectors, plugin_printers



from typing import TYPE_CHECKING
from slither.slithir.operations.lvalue import OperationWithLValue
from slither.slithir.utils.utils import is_valid_lvalue
from slither.core.declarations.solidity_variables import SolidityVariableComposed
from slither.slithir.variables.local_variable import LocalIRVariable
from slither.slithir.variables.state_variable import StateIRVariable
from slither.slithir.variables.temporary_ssa import TemporaryVariableSSA

if TYPE_CHECKING:
    from slither.core.cfg.node import Node


class Phi(OperationWithLValue):
    def __init__(
        self, left_variable: LocalIRVariable | StateIRVariable, nodes: set["Node"]
    ) -> None:
        # When Phi operations are created the
        # correct indexes of the variables are not yet computed
        # We store the nodes where the variables are written
        # so we can update the rvalues of the Phi operation
        # after its instantiation
        assert is_valid_lvalue(left_variable)
        assert isinstance(nodes, set)
        super().__init__()
        self._lvalue = left_variable
        self._rvalues = []
        self._nodes = nodes

    @property
    def read(
        self,
    ) -> list[SolidityVariableComposed | LocalIRVariable | TemporaryVariableSSA | StateIRVariable]:
        return self.rvalues

    @property
    def rvalues(self):
        return self._rvalues

    @rvalues.setter
    def rvalues(self, vals):
        self._rvalues = vals

    @property
    def nodes(self) -> set["Node"]:
        return self._nodes

    def __str__(self):
        return f"{self.lvalue}({self.lvalue.type}) := \u03d5({[str(v) for v in self._rvalues]})"



from collections import defaultdict

from slither.analyses.data_dependency.data_dependency import is_dependent
from slither.core.cfg.node import Node
from slither.core.declarations import Function, Contract, SolidityVariableComposed
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations import LowLevelCall, HighLevelCall
from slither.utils.output import Output


def _detect_token_reentrant(contract: Contract) -> dict[Function, list[Node]]:
    ret: dict[Function, list[Node]] = defaultdict(list)
    for function in contract.functions_entry_points:
        if function.full_name in [
            "transfer(address,uint256)",
            "transferFrom(address,address,uint256)",
        ]:
            for ir in function.all_slithir_operations():
                if isinstance(ir, (LowLevelCall, HighLevelCall)):
                    if not function.parameters:
                        continue
                    if any(
                        is_dependent(ir.destination, parameter, function)
                        for parameter in function.parameters
                    ):
                        ret[function].append(ir.node)
                    if is_dependent(
                        ir.destination, SolidityVariableComposed("msg.sender"), function
                    ):
                        ret[function].append(ir.node)
                    if is_dependent(
                        ir.destination, SolidityVariableComposed("tx.origin"), function
                    ):
                        ret[function].append(ir.node)
    return ret


class TokenReentrancy(AbstractDetector):
    ARGUMENT = "token-reentrancy"
    HELP = "Tokens that are reentrancies unsafe"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#token-reentrant"

    WIKI_TITLE = "Token reentrant"

    # region wiki_description
    WIKI_DESCRIPTION = """
    Tokens that allow arbitrary external call on transfer/transfer (such as ERC223/ERC777) can be exploited on third
    party through a reentrancy."""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
    ```solidity
contract MyToken{
    function transferFrom(address from, address to, uint) public{
        // do some stuff
        from.call("..")
        // do some stuff
    }
}

contract MyDefi{
    function convert(ERC token) public{
        // do some stuff
        token.transferFrom(..)
        //
    }
}
    ```

    `MyDefi` has a reentrancy, but its developers did not think transferFrom could be reentrancy.
    `MyToken` is used in MyDefi. As a result an attacker can exploit the reentrancy."""
    # endregion wiki_exploit_scenario

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """Avoid to have external calls in `transfer`/`transferFrom`.
If you do, ensure your users are aware of the potential issues."""
    # endregion wiki_recommendation

    def _detect(self) -> list[Output]:
        results = []
        for contract in self.compilation_unit.contracts_derived:
            vulns = _detect_token_reentrant(contract)
            for function, nodes in vulns.items():
                info: DETECTOR_INFO = [function, " is an reentrancy unsafe token function:\n"]
                for node in nodes:
                    info += ["\t-", node, "\n"]
                json = self.generate_result(info)
                results.append(json)

        return results



import os
from pathlib import Path

from slither import Slither
from slither.utils.code_generation import (
    generate_interface,
)

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data" / "code_generation"


def test_interface_generation(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.4")

    sl = Slither(Path(TEST_DATA_DIR, "CodeGeneration.sol").as_posix(), solc=solc_path)

    actual = generate_interface(sl.get_contract_from_name("TestContract")[0])
    expected_path = Path(TEST_DATA_DIR, "TEST_generated_code.sol").as_posix()

    with open(expected_path, encoding="utf-8") as file:
        expected = file.read()

    assert actual == expected

    actual = generate_interface(sl.get_contract_from_name("TestContract")[0], unroll_structs=False)
    expected_path = os.path.join(TEST_DATA_DIR, "TEST_generated_code_not_unrolled.sol")

    with open(expected_path, encoding="utf-8") as file:
        expected = file.read()

    assert actual == expected



import sys

import logging
from argparse import ArgumentParser, Namespace

from crytic_compile import cryticparser
from slither import Slither
from slither.core.declarations import FunctionContract
from slither.utils.colors import red
from slither.tools.possible_paths.possible_paths import (
    find_target_paths,
    resolve_functions,
    ResolveFunctionException,
)

logging.basicConfig()
logging.getLogger("Slither").setLevel(logging.INFO)


def parse_args() -> Namespace:
    """
    Parse the underlying arguments for the program.
    :return: Returns the arguments for the program.
    """
    parser: ArgumentParser = ArgumentParser(
        description="PossiblePaths",
        usage="possible_paths.py filename [contract.function targets]",
    )

    parser.add_argument(
        "filename", help="The filename of the contract or truffle directory to analyze."
    )

    parser.add_argument("targets", nargs="+")

    cryticparser.init(parser)

    return parser.parse_args()


def main() -> None:
    # ------------------------------
    # PossiblePaths.py
    #       Usage: python3 possible_paths.py filename targets
    #       Example: python3 possible_paths.py contract.sol contract1.function1 contract2.function2 contract3.function3
    # ------------------------------
    # Parse all arguments
    args = parse_args()

    # Perform slither analysis on the given filename
    slither = Slither(args.filename, **vars(args))

    try:
        targets = resolve_functions(slither, args.targets)
    except ResolveFunctionException as resolvefunction:
        print(red(resolvefunction))
        sys.exit(-1)

    # Print out all target functions.
    print("Target functions:")
    for target in targets:
        if isinstance(target, FunctionContract):
            print(f"- {target.contract_declarer.name}.{target.full_name}")
        else:
            pass
            # TODO implement me
    print("\n")

    # Obtain all paths which reach the target functions.
    reaching_paths = find_target_paths(slither, targets)
    reaching_functions = {y for x in reaching_paths for y in x if y not in targets}

    # Print out all function names which can reach the targets.
    print("The following functions reach the specified targets:")
    for function_desc in sorted([f"{f.canonical_name}" for f in reaching_functions]):
        print(f"- {function_desc}")
    print("\n")

    # Format all function paths.
    reaching_paths_str = [
        " -> ".join([f"{f.canonical_name}" for f in reaching_path])
        for reaching_path in reaching_paths
    ]

    # Print a sorted list of all function paths which can reach the targets.
    print("The following paths reach the specified targets:")
    for reaching_path in sorted(reaching_paths_str):
        print(f"{reaching_path}\n")


if __name__ == "__main__":
    main()



"""
Variable module
"""

from typing import Optional, TYPE_CHECKING, Union

from slither.core.source_mapping.source_mapping import SourceMapping
from slither.core.solidity_types.type import Type
from slither.core.solidity_types.elementary_type import ElementaryType

if TYPE_CHECKING:
    from slither.core.expressions.expression import Expression
    from slither.core.declarations import Function


class Variable(SourceMapping):
    def __init__(self) -> None:
        super().__init__()
        self._name: str | None = None
        self._initial_expression: Expression | None = None
        self._type: list | Type | Function | str | None = None
        self._initialized: bool | None = None
        self._visibility: str | None = None
        self._is_constant = False
        self._is_immutable: bool = False
        self._is_reentrant: bool = True
        self._write_protection: list[str] | None = None

    @property
    def is_scalar(self) -> bool:
        return isinstance(self.type, ElementaryType)

    @property
    def expression(self) -> Optional["Expression"]:
        """
        Expression: Expression of the node (if initialized)
        Initial expression may be different than the expression of the node
        where the variable is declared, if its used ternary operator
        Ex: uint a = b?1:2
        The expression associated to a is uint a = b?1:2
        But two nodes are created,
        one where uint a = 1,
        and one where uint a = 2

        """
        return self._initial_expression

    @expression.setter
    def expression(self, expr: "Expression") -> None:
        self._initial_expression = expr

    @property
    def initialized(self) -> bool | None:
        """
        boolean: True if the variable is initialized at construction
        """
        return self._initialized

    @initialized.setter
    def initialized(self, is_init: bool) -> None:
        self._initialized = is_init

    @property
    def uninitialized(self) -> bool:
        """
        boolean: True if the variable is not initialized
        """
        return not self._initialized

    @property
    def name(self) -> str | None:
        """
        str: variable name
        """
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = name

    @property
    def type(self) -> Union[list, Type, "Function", str] | None:
        return self._type

    @type.setter
    def type(self, new_type: Type) -> None:
        assert isinstance(new_type, Type)
        self._type = new_type

    @property
    def is_constant(self) -> bool:
        return self._is_constant

    @is_constant.setter
    def is_constant(self, is_cst: bool) -> None:
        self._is_constant = is_cst

    @property
    def is_reentrant(self) -> bool:
        return self._is_reentrant

    @is_reentrant.setter
    def is_reentrant(self, is_reentrant: bool) -> None:
        self._is_reentrant = is_reentrant

    @property
    def write_protection(self) -> list[str] | None:
        return self._write_protection

    @write_protection.setter
    def write_protection(self, write_protection: list[str]) -> None:
        self._write_protection = write_protection

    @property
    def visibility(self) -> str | None:
        """
        str: variable visibility
        """
        return self._visibility

    @visibility.setter
    def visibility(self, v: str) -> None:
        self._visibility = v

    def set_type(self, t: Union[list, Type, "Function", str] | None) -> None:
        if isinstance(t, str):
            self._type = ElementaryType(t)
            return
        assert isinstance(t, (Type, list)) or t is None
        self._type = t

    @property
    def is_immutable(self) -> bool:
        """
        Return true of the variable is immutable

        :return:
        """
        return self._is_immutable

    @is_immutable.setter
    def is_immutable(self, immutablility: bool) -> None:
        self._is_immutable = immutablility

    ###################################################################################
    ###################################################################################
    # region Signature
    ###################################################################################
    ###################################################################################

    @property
    def signature(self) -> tuple[str, list[str], list[str]]:
        """
        Return the signature of the state variable as a function signature
        :return: (str, list(str), list(str)), as (name, list parameters type, list return values type)
        """

        from slither.utils.type import (
            export_nested_types_from_variable,
            export_return_type_from_variable,
        )

        return (
            self.name,
            [str(x) for x in export_nested_types_from_variable(self)],  # type: ignore
            [str(x) for x in export_return_type_from_variable(self)],  # type: ignore
        )

    @property
    def signature_str(self) -> str:
        """
        Return the signature of the state variable as a function signature
        :return: str: func_name(type1,type2) returns(type3)
        """
        name, parameters, returnVars = self.signature
        return name + "(" + ",".join(parameters) + ") returns(" + ",".join(returnVars) + ")"

    @property
    def solidity_signature(self) -> str:
        name, parameters, _ = self.signature
        return f"{name}({','.join(parameters)})"

    def __str__(self) -> str:
        if self._name is None:
            return ""
        return self._name



from slither.core.cfg.node import NodeType, Node
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations import LowLevelCall, InternalCall
from slither.core.declarations import Contract
from slither.utils.output import Output

Result = list[tuple[Node, list[str]]]


def detect_delegatecall_in_loop(contract: Contract) -> Result:
    results: Result = []
    for f in contract.functions_entry_points:
        if f.is_implemented and f.payable:
            delegatecall_in_loop(f.entry_point, 0, [], [], results)
    return results


def delegatecall_in_loop(
    node: Node | None,
    in_loop_counter: int,
    visited: list[Node],
    calls_stack: list[str],
    results: Result,
) -> None:
    if node is None:
        return

    if node in visited:
        return
    # shared visited
    visited.append(node)

    if node.type == NodeType.STARTLOOP:
        in_loop_counter += 1
    elif node.type == NodeType.ENDLOOP:
        in_loop_counter -= 1

    for ir in node.irs:
        if (
            in_loop_counter > 0
            and isinstance(ir, (LowLevelCall))
            and ir.function_name == "delegatecall"
        ):
            results.append((ir.node, calls_stack.copy()))
        if isinstance(ir, (InternalCall)) and ir.function:
            calls_stack.append(node.function.canonical_name)
            delegatecall_in_loop(
                ir.function.entry_point, in_loop_counter, visited, calls_stack, results
            )
            calls_stack.pop()

    for son in node.sons:
        delegatecall_in_loop(son, in_loop_counter, visited, calls_stack, results)


class DelegatecallInLoop(AbstractDetector):
    """
    Detect the use of delegatecall inside a loop in a payable function
    """

    ARGUMENT = "delegatecall-loop"
    HELP = "Payable functions using `delegatecall` inside a loop"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation/#payable-functions-using-delegatecall-inside-a-loop"

    WIKI_TITLE = "Payable functions using `delegatecall` inside a loop"
    WIKI_DESCRIPTION = "Detect the use of `delegatecall` inside a loop in a payable function."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract DelegatecallInLoop{

    mapping (address => uint256) balances;

    function bad(address[] memory receivers) public payable {
        for (uint256 i = 0; i < receivers.length; i++) {
            address(this).delegatecall(abi.encodeWithSignature("addBalance(address)", receivers[i]));
        }
    }

    function addBalance(address a) public payable {
        balances[a] += msg.value;
    }

}
```
When calling `bad` the same `msg.value` amount will be accredited multiple times."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = """
Carefully check that the function called by `delegatecall` is not payable/doesn't use `msg.value`.
"""

    def _detect(self) -> list[Output]:
        """"""
        results: list[Output] = []
        for c in self.compilation_unit.contracts_derived:
            values = detect_delegatecall_in_loop(c)
            for node, calls_stack in values:
                func = node.function

                info: DETECTOR_INFO = [
                    func,
                    " has delegatecall inside a loop in a payable function: ",
                    node,
                    "\n",
                ]

                if len(calls_stack) > 0:
                    info.append("\tCalls stack containing the loop:\n")
                    for call in calls_stack:
                        info.extend(["\t\t", call, "\n"])

                res = self.generate_result(info)
                results.append(res)

        return results



from typing import Optional, TYPE_CHECKING

from slither.core.declarations.contract_level import ContractLevel
from slither.core.variables.variable import Variable

if TYPE_CHECKING:
    from slither.core.cfg.node import Node
    from slither.core.declarations import Contract


class StateVariable(ContractLevel, Variable):
    def __init__(self) -> None:
        super().__init__()
        self._node_initialization: Node | None = None
        self._location: str | None = None

    def is_declared_by(self, contract: "Contract") -> bool:
        """
        Check if the element is declared by the contract
        :param contract:
        :return:
        """
        return self.contract == contract

    def set_location(self, loc: str) -> None:
        self._location = loc

    @property
    def location(self) -> str | None:
        """
            Variable Location
            Can be default or transient
        Returns:
            (str)
        """
        return self._location

    @property
    def is_stored(self) -> bool:
        """
        Checks if the state variable is stored, based on it not being constant or immutable or transient.
        """
        return not self._is_constant and not self._is_immutable and self._location != "transient"

    @property
    def is_transient(self) -> bool:
        """
        Checks if the state variable is transient. A transient variable can not be constant or immutable.
        """
        return self._location == "transient"

    # endregion
    ###################################################################################
    ###################################################################################
    # region Name
    ###################################################################################
    ###################################################################################

    @property
    def canonical_name(self) -> str:
        return f"{self.contract.name}.{self.name}"

    @property
    def full_name(self) -> str:
        """
        Return the name of the state variable as a function signaure
        str: func_name(type1,type2)
        :return: the function signature without the return values
        """
        name, parameters, _ = self.signature
        return name + "(" + ",".join(parameters) + ")"

    # endregion
    ###################################################################################
    ###################################################################################
    # region IRs (initialization)
    ###################################################################################
    ###################################################################################

    @property
    def node_initialization(self) -> Optional["Node"]:
        """
        Node for the state variable initalization
        :return:
        """
        return self._node_initialization

    @node_initialization.setter
    def node_initialization(self, node_initialization):
        self._node_initialization = node_initialization

    # endregion
    ###################################################################################
    ###################################################################################



import math

from slither.core.solidity_types import (
    ArrayType,
    MappingType,
    ElementaryType,
    UserDefinedType,
    TypeAlias,
)
from slither.core.solidity_types.type import Type
from slither.core.variables.variable import Variable


def _convert_type_for_solidity_signature_to_string(
    types: Type | list[Type], seen: set[Type]
) -> str:
    if isinstance(types, Type):
        # Array might be struct, so we need to go again through the conversion here
        # We could have this logic in convert_type_for_solidity_signature
        # But the slither type system is not straightforward to manipulate here
        # And it would require to create a new ArrayType, which a potential List[Type] as input
        # Which is currently not supported. This comes down to (uint, uint)[] not being possible in Solidity
        # While having an array of a struct of two uint leads to a (uint, uint)[] signature
        if isinstance(types, ArrayType):
            underlying_type = convert_type_for_solidity_signature(types.type, seen)
            underlying_type_str = _convert_type_for_solidity_signature_to_string(
                underlying_type, seen
            )

            if types.length is None:
                return underlying_type_str + "[]"
            return underlying_type_str + f"[{types.length_value}]"

        return str(types)

    first_item = True

    ret = "("
    for underlying_type in types:
        if first_item:
            ret += _convert_type_for_solidity_signature_to_string(underlying_type, seen)
        else:
            ret += "," + _convert_type_for_solidity_signature_to_string(underlying_type, seen)
        first_item = False

    ret += ")"
    return ret


def convert_type_for_solidity_signature_to_string(t: Type) -> str:
    seen: set[Type] = set()
    types = convert_type_for_solidity_signature(t, seen)
    return _convert_type_for_solidity_signature_to_string(types, seen)


def convert_type_for_solidity_signature(t: Type, seen: set[Type]) -> Type | list[Type]:
    from slither.core.declarations import Contract, Enum, Structure

    # Solidity allows recursive type for structure definition if its not used in public /external
    # When this happens we can reach an infinite loop. If we detect a loop, we just stop converting the underlying type
    # This is ok, because it wont happen for public/external function
    #
    # contract A{
    #
    #   struct St{
    #       St[] a;
    #       uint b;
    #   }
    #
    #    function f(St memory s) internal{}
    #
    # }
    if t in seen:
        return t
    seen.add(t)

    if isinstance(t, UserDefinedType):
        underlying_type = t.type
        if isinstance(underlying_type, Contract):
            return ElementaryType("address")
        if isinstance(underlying_type, Enum):
            number_values = len(underlying_type.values)
            # IF below 65536, avoid calling log2
            if number_values <= 256:
                uint = "8"
            elif number_values <= 65536:
                uint = "16"
            else:
                uint = str(int(math.log2(number_values)))
            return ElementaryType(f"uint{uint}")
        if isinstance(underlying_type, Structure):
            # We can't have recursive types for structure, so recursion is ok here
            types = [
                convert_type_for_solidity_signature(x.type, seen)
                for x in underlying_type.elems_ordered
            ]
            return types

    if isinstance(t, TypeAlias):
        return t.type

    return t


def _export_nested_types_from_variable(
    current_type: Type, ret: list[Type], seen: set[Type]
) -> None:
    """
    Export the list of nested types (mapping/array)
    :param variable:
    :return: list(Type)
    """
    if isinstance(current_type, MappingType):
        underlying_type = convert_type_for_solidity_signature(current_type.type_from, seen)
        if isinstance(underlying_type, list):
            ret.extend(underlying_type)
        else:
            ret.append(underlying_type)
        next_type = current_type.type_to

    elif isinstance(current_type, ArrayType):
        ret.append(ElementaryType("uint256"))
        next_type = current_type.type
    else:
        return
    _export_nested_types_from_variable(next_type, ret, seen)


def export_nested_types_from_variable(variable: Variable) -> list[Type]:
    """
    Export the list of nested types (mapping/array)
    :param variable:
    :return: list(Type)
    """
    l: list[Type] = []
    seen: set[Type] = set()
    _export_nested_types_from_variable(variable.type, l, seen)
    return l


def _export_return_type_from_variable(underlying_type: Type, all_types: bool) -> list[Type]:
    from slither.core.declarations import Structure

    if isinstance(underlying_type, MappingType):
        if not all_types:
            return []
        return export_return_type_from_variable(underlying_type.type_to)

    if isinstance(underlying_type, ArrayType):
        if not all_types:
            return []
        return export_return_type_from_variable(underlying_type.type)

    if isinstance(underlying_type, UserDefinedType) and isinstance(underlying_type.type, Structure):
        ret = []
        for r in underlying_type.type.elems_ordered:
            ret += export_return_type_from_variable(r, all_types=False)

        return ret

    return [underlying_type]


def export_return_type_from_variable(
    variable_or_type: Type | Variable, all_types: bool = True
) -> list[Type]:
    """
    Return the type returned by a variable.
    If all_types set to false, filter array/mapping. This is useful as the mapping/array in a structure are not
    returned by solidity
    :param variable_or_type
    :param all_types
    :return: Type
    """

    from slither.core.declarations import Structure

    if isinstance(variable_or_type, Type):
        return _export_return_type_from_variable(variable_or_type, all_types)

    if isinstance(variable_or_type.type, MappingType):
        if not all_types:
            return []
        return export_return_type_from_variable(variable_or_type.type.type_to)

    if isinstance(variable_or_type.type, ArrayType):
        if not all_types:
            return []
        return export_return_type_from_variable(variable_or_type.type.type)

    if isinstance(variable_or_type.type, UserDefinedType) and isinstance(
        variable_or_type.type.type, Structure
    ):
        ret = []
        for r in variable_or_type.type.type.elems_ordered:
            ret += export_return_type_from_variable(r, all_types=False)
        return ret

    return [variable_or_type.type]


def is_underlying_type_address(t: "Type") -> bool:
    """
    Return true if the underlying type is an address
    i.e. if the type is an address or a contract
    """

    from slither.core.declarations.contract import Contract

    if t == ElementaryType("address"):
        return True
    if isinstance(t, UserDefinedType) and isinstance(t.type, Contract):
        return True
    return False



"""
Module detecting deprecated standards.
"""

from slither.core.cfg.node import Node, NodeType
from slither.core.declarations.contract import Contract
from slither.core.declarations.solidity_variables import (
    SolidityVariableComposed,
    SolidityFunction,
)
from slither.core.expressions.expression import Expression
from slither.core.variables import StateVariable
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations import LowLevelCall
from slither.utils.output import Output
from slither.visitors.expression.export_values import ExportValues


# Reference: https://smartcontractsecurity.github.io/SWC-registry/docs/SWC-111
class DeprecatedStandards(AbstractDetector):
    """
    Use of Deprecated Standards
    """

    ARGUMENT = "deprecated-standards"
    HELP = "Deprecated Solidity Standards"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.HIGH
    LANGUAGE = "solidity"
    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#deprecated-standards"

    WIKI_TITLE = "Deprecated standards"
    WIKI_DESCRIPTION = "Detect the usage of deprecated standards."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract ContractWithDeprecatedReferences {
    // Deprecated: Change block.blockhash() -> blockhash()
    bytes32 globalBlockHash = block.blockhash(0);

    // Deprecated: Change constant -> view
    function functionWithDeprecatedThrow() public constant {
        // Deprecated: Change msg.gas -> gasleft()
        if(msg.gas == msg.value) {
            // Deprecated: Change throw -> revert()
            throw;
        }
    }

    // Deprecated: Change constant -> view
    function functionWithDeprecatedReferences() public constant {
        // Deprecated: Change sha3() -> keccak256()
        bytes32 sha3Result = sha3("test deprecated sha3 usage");

        // Deprecated: Change callcode() -> delegatecall()
        address(this).callcode();

        // Deprecated: Change suicide() -> selfdestruct()
        suicide(address(0));
    }
}
```"""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Replace all uses of deprecated symbols."

    # The format for the following deprecated lists is [(detecting_signature, original_text, recommended_text)]
    DEPRECATED_SOLIDITY_VARIABLE = [
        ("block.blockhash", "block.blockhash()", "blockhash()"),
        ("msg.gas", "msg.gas", "gasleft()"),
    ]
    DEPRECATED_SOLIDITY_FUNCTIONS = [
        ("suicide(address)", "suicide()", "selfdestruct()"),
        ("sha3()", "sha3()", "keccak256()"),
    ]
    DEPRECATED_NODE_TYPES = [(NodeType.THROW, "throw", "revert()")]
    DEPRECATED_LOW_LEVEL_CALLS = [("callcode", "callcode", "delegatecall")]

    def detect_deprecation_in_expression(
        self, expression: Expression
    ) -> list[tuple[str, str, str]]:
        """Detects if an expression makes use of any deprecated standards.

        Returns:
            list of tuple: (detecting_signature, original_text, recommended_text)"""
        # Perform analysis on this expression
        export = ExportValues(expression)
        export_values = export.result()

        # Define our results list
        results = []

        # Check if there is usage of any deprecated solidity variables or functions
        for dep_var in self.DEPRECATED_SOLIDITY_VARIABLE:
            if SolidityVariableComposed(dep_var[0]) in export_values:
                results.append(dep_var)
        for dep_func in self.DEPRECATED_SOLIDITY_FUNCTIONS:
            if SolidityFunction(dep_func[0]) in export_values:
                results.append(dep_func)

        return results

    def detect_deprecated_references_in_node(
        self, node: Node
    ) -> list[tuple[str | NodeType, str, str]]:
        """Detects if a node makes use of any deprecated standards.

        Returns:
            list of tuple: (detecting_signature, original_text, recommended_text)"""
        # Define our results list
        results: list[tuple[str | NodeType, str, str]] = []

        # If this node has an expression, we check the underlying expression.
        if node.expression:
            results += self.detect_deprecation_in_expression(node.expression)

        # Check if there is usage of any deprecated solidity variables or functions
        for dep_node in self.DEPRECATED_NODE_TYPES:
            if node.type == dep_node[0]:
                results.append(dep_node)

        return results

    def detect_deprecated_references_in_contract(
        self, contract: Contract
    ) -> list[
        tuple[StateVariable, list[tuple[str, str, str]]]
        | tuple[Node, list[tuple[str | NodeType, str, str]]]
    ]:
        """Detects the usage of any deprecated built-in symbols.

        Returns:
            list of tuple: (state_variable | node, (detecting_signature, original_text, recommended_text))"""
        results: list[
            tuple[StateVariable, list[tuple[str, str, str]]]
            | tuple[Node, list[tuple[str | NodeType, str, str]]]
        ] = []

        for state_variable in contract.state_variables_declared:
            if state_variable.expression:
                deprecated_results = self.detect_deprecation_in_expression(
                    state_variable.expression
                )
                if deprecated_results:
                    results.append((state_variable, deprecated_results))

        # Loop through all functions + modifiers in this contract.

        for function in contract.functions_and_modifiers_declared:
            # Loop through each node in this function.
            for node in function.nodes:
                # Detect deprecated references in the node.
                deprecated_results_node = self.detect_deprecated_references_in_node(node)

                # Detect additional deprecated low-level-calls.
                for ir in node.irs:
                    if isinstance(ir, LowLevelCall):
                        for dep_llc in self.DEPRECATED_LOW_LEVEL_CALLS:
                            if ir.function_name == dep_llc[0]:
                                deprecated_results_node.append(dep_llc)

                # If we have any results from this iteration, add them to our results list.
                if deprecated_results_node:
                    results.append((node, deprecated_results_node))

        return results

    def _detect(self) -> list[Output]:
        """Detects if an expression makes use of any deprecated standards.

        Recursively visit the calls
        Returns:
            list: {'vuln', 'filename,'contract','func', 'deprecated_references'}

        """
        results = []
        for contract in self.contracts:
            deprecated_references = self.detect_deprecated_references_in_contract(contract)
            if deprecated_references:
                for deprecated_reference in deprecated_references:
                    source_object = deprecated_reference[0]
                    deprecated_entries = deprecated_reference[1]
                    info: DETECTOR_INFO = ["Deprecated standard detected ", source_object, ":\n"]

                    for _dep_id, original_desc, recommended_disc in deprecated_entries:
                        info += [
                            f'\t- Usage of "{original_desc}" should be replaced with "{recommended_disc}"\n'
                        ]

                    res = self.generate_result(info)
                    results.append(res)

        return results



from collections import namedtuple

ERC = namedtuple("ERC", ["name", "parameters", "return_type", "view", "required", "events"])
ERC_EVENT = namedtuple("ERC_EVENT", ["name", "parameters", "indexes"])


def erc_to_signatures(erc: list[ERC]):
    """
    Return the list of mandatory signatures
    :param erc:
    :return:
    """
    return [f"{e.name}({','.join(e.parameters)})" for e in erc if e.required]


# Final
# https://eips.ethereum.org/EIPS/eip-20

ERC20_transfer_event = ERC_EVENT("Transfer", ["address", "address", "uint256"], [True, True, False])
ERC20_approval_event = ERC_EVENT("Approval", ["address", "address", "uint256"], [True, True, False])
ERC20_EVENTS = [ERC20_transfer_event, ERC20_approval_event]

ERC20 = [
    ERC("totalSupply", [], "uint256", True, True, []),
    ERC("balanceOf", ["address"], "uint256", True, True, []),
    ERC("transfer", ["address", "uint256"], "bool", False, True, [ERC20_transfer_event]),
    ERC(
        "transferFrom",
        ["address", "address", "uint256"],
        "bool",
        False,
        True,
        [ERC20_transfer_event],
    ),
    ERC("approve", ["address", "uint256"], "bool", False, True, [ERC20_approval_event]),
    ERC("allowance", ["address", "address"], "uint256", True, True, []),
]

ERC20_OPTIONAL = [
    ERC("name", [], "string", True, False, []),
    ERC("symbol", [], "string", True, False, []),
    ERC("decimals", [], "uint8", True, False, []),
]

ERC20 = ERC20 + ERC20_OPTIONAL

ERC20_signatures = erc_to_signatures(ERC20)

# Draft
# https://github.com/ethereum/eips/issues/223

ERC223_transfer_event = ERC_EVENT(
    "Transfer", ["address", "address", "uint256", "bytes"], [True, True, False, False]
)

ERC223_EVENTS = [ERC223_transfer_event]

ERC223 = [
    ERC("name", [], "string", True, True, []),
    ERC("symbol", [], "string", True, True, []),
    ERC("decimals", [], "uint8", True, True, []),
    ERC("totalSupply", [], "uint256", True, True, []),
    ERC("balanceOf", ["address"], "uint256", True, True, []),
    ERC("transfer", ["address", "uint256"], "bool", False, True, [ERC223_transfer_event]),
    ERC(
        "transfer",
        ["address", "uint256", "bytes"],
        "bool",
        False,
        True,
        [ERC223_transfer_event],
    ),
    ERC(
        "transfer",
        ["address", "uint256", "bytes", "string"],
        "bool",
        False,
        True,
        [ERC223_transfer_event],
    ),
]
ERC223_signatures = erc_to_signatures(ERC223)

# Final
# https://eips.ethereum.org/EIPS/eip-165

ERC165_EVENTS: list = []

ERC165 = [ERC("supportsInterface", ["bytes4"], "bool", True, True, [])]
ERC165_signatures = erc_to_signatures(ERC165)

# Final
# https://eips.ethereum.org/EIPS/eip-721
# Must have ERC165

ERC721_transfer_event = ERC_EVENT("Transfer", ["address", "address", "uint256"], [True, True, True])
ERC721_approval_event = ERC_EVENT("Approval", ["address", "address", "uint256"], [True, True, True])
ERC721_approvalforall_event = ERC_EVENT(
    "ApprovalForAll", ["address", "address", "bool"], [True, True, False]
)
ERC721_EVENTS = [
    ERC721_transfer_event,
    ERC721_approval_event,
    ERC721_approvalforall_event,
]

ERC721 = [
    ERC("balanceOf", ["address"], "uint256", True, True, []),
    ERC("ownerOf", ["uint256"], "address", True, True, []),
    ERC(
        "safeTransferFrom",
        ["address", "address", "uint256", "bytes"],
        "",
        False,
        True,
        [ERC721_transfer_event],
    ),
    ERC(
        "safeTransferFrom",
        ["address", "address", "uint256"],
        "",
        False,
        True,
        [ERC721_transfer_event],
    ),
    ERC(
        "transferFrom",
        ["address", "address", "uint256"],
        "",
        False,
        True,
        [ERC721_transfer_event],
    ),
    ERC("approve", ["address", "uint256"], "", False, True, [ERC721_approval_event]),
    ERC(
        "setApprovalForAll",
        ["address", "bool"],
        "",
        False,
        True,
        [ERC721_approvalforall_event],
    ),
    ERC("getApproved", ["uint256"], "address", True, True, []),
    ERC("isApprovedForAll", ["address", "address"], "bool", True, True, []),
] + ERC165

ERC721_OPTIONAL = [
    ERC("name", [], "string", True, False, []),
    ERC("symbol", [], "string", False, False, []),
    ERC("tokenURI", ["uint256"], "string", False, False, []),
]

ERC721 = ERC721 + ERC721_OPTIONAL

ERC721_signatures = erc_to_signatures(ERC721)

# Final
# https://eips.ethereum.org/EIPS/eip-1820
ERC1820_EVENTS: list = []
ERC1820 = [
    ERC(
        "canImplementInterfaceForAddress",
        ["bytes32", "address"],
        "bytes32",
        True,
        True,
        [],
    )
]
ERC1820_signatures = erc_to_signatures(ERC1820)

# Last Call
# https://eips.ethereum.org/EIPS/eip-777
ERC777_sent_event = ERC_EVENT(
    "Sent",
    ["address", "address", "address", "uint256", "bytes", "bytes"],
    [True, True, True, False, False, False],
)
ERC777_minted_event = ERC_EVENT(
    "Minted",
    ["address", "address", "uint256", "bytes", "bytes"],
    [True, True, False, False, False],
)
ERC777_burned_event = ERC_EVENT(
    "Burned",
    ["address", "address", "uint256", "bytes", "bytes"],
    [True, True, False, False, False],
)
ERC777_authorizedOperator_event = ERC_EVENT(
    "AuthorizedOperator", ["address", "address"], [True, True]
)
ERC777_revokedoperator_event = ERC_EVENT("RevokedOperator", ["address", "address"], [True, True])
ERC777_EVENTS = [
    ERC777_sent_event,
    ERC777_minted_event,
    ERC777_burned_event,
    ERC777_authorizedOperator_event,
    ERC777_revokedoperator_event,
]

ERC777 = [
    ERC("name", [], "string", True, True, []),
    ERC("symbol", [], "string", True, True, []),
    ERC("totalSupply", [], "uint256", True, True, []),
    ERC("balanceOf", ["address"], "uint256", True, True, []),
    ERC("granularity", [], "uint256", True, True, []),
    ERC("defaultOperators", [], "address[]", True, True, []),
    ERC("isOperatorFor", ["address", "address"], "bool", True, True, []),
    ERC(
        "authorizeOperator",
        ["address"],
        "",
        False,
        True,
        [ERC777_authorizedOperator_event],
    ),
    ERC("revokeOperator", ["address"], "", False, True, [ERC777_revokedoperator_event]),
    ERC("send", ["address", "uint256", "bytes"], "", False, True, [ERC777_sent_event]),
    ERC(
        "operatorSend",
        ["address", "address", "uint256", "bytes", "bytes"],
        "",
        False,
        True,
        [ERC777_sent_event],
    ),
    ERC("burn", ["uint256", "bytes"], "", False, True, [ERC777_burned_event]),
    ERC(
        "operatorBurn",
        ["address", "uint256", "bytes", "bytes"],
        "",
        False,
        True,
        [ERC777_burned_event],
    ),
]
ERC777_signatures = erc_to_signatures(ERC777)

# Final
# https://eips.ethereum.org/EIPS/eip-1155
# Must have ERC165

ERC1155_transfersingle_event = ERC_EVENT(
    "TransferSingle",
    ["address", "address", "address", "uint256", "uint256"],
    [True, True, True, False, False],
)

ERC1155_transferbatch_event = ERC_EVENT(
    "TransferBatch",
    ["address", "address", "address", "uint256[]", "uint256[]"],
    [True, True, True, False, False],
)

ERC1155_approvalforall_event = ERC_EVENT(
    "ApprovalForAll",
    ["address", "address", "bool"],
    [True, True, False],
)

ERC1155_uri_event = ERC_EVENT(
    "URI",
    ["string", "uint256"],
    [False, True],
)

ERC1155_EVENTS = [
    ERC1155_transfersingle_event,
    ERC1155_transferbatch_event,
    ERC1155_approvalforall_event,
    ERC1155_uri_event,
]

ERC1155 = [
    ERC(
        "safeTransferFrom",
        ["address", "address", "uint256", "uint256", "bytes"],
        "",
        False,
        True,
        [ERC1155_transfersingle_event],
    ),
    ERC(
        "safeBatchTransferFrom",
        ["address", "address", "uint256[]", "uint256[]", "bytes"],
        "",
        False,
        True,
        [],
    ),
    ERC("balanceOf", ["address", "uint256"], "uint256", True, True, []),
    ERC("balanceOfBatch", ["address[]", "uint256[]"], "uint256[]", True, True, []),
    ERC("setApprovalForAll", ["address", "bool"], "", False, True, [ERC1155_approvalforall_event]),
    ERC("isApprovedForAll", ["address", "address"], "bool", True, True, []),
] + ERC165

ERC1155_TOKEN_RECEIVER = [
    ERC(
        "onERC1155Received",
        ["address", "address", "uint256", "uint256", "bytes"],
        "bytes4",
        False,
        False,
        [],
    ),
    ERC(
        "onERC1155BatchReceived",
        ["address", "address", "uint256[]", "uint256[]", "bytes"],
        "bytes4",
        False,
        False,
        [],
    ),
]

ERC1155_METADATA = [ERC("uri", ["uint256"], "string", True, False, [])]

ERC1155 = ERC1155 + ERC1155_TOKEN_RECEIVER + ERC1155_METADATA

ERC1155_signatures = erc_to_signatures(ERC1155)

# Review
# https://eips.ethereum.org/EIPS/eip-2612
# Must have ERC20

ERC2612_EVENTS = []
ERC2612 = [
    ERC(
        "permit",
        ["address", "address", "uint256", "uint256", "uint8", "bytes32", "bytes32"],
        "",
        False,
        True,
        [],
    ),
    ERC("nonces", ["address"], "uint256", True, True, []),
    ERC("DOMAIN_SEPARATOR", [], "bytes32", True, True, []),
] + ERC20

ERC2612_signatures = erc_to_signatures(ERC2612)

# Review
# https://eips.ethereum.org/EIPS/eip-1363
# Must have ERC20 and ERC165

ERC1363_EVENTS = []
ERC1363 = (
    [
        ERC("transferAndCall", ["address", "uint256"], "bool", False, True, []),
        ERC("transferAndCall", ["address", "uint256", "bytes"], "bool", False, True, []),
        ERC("transferFromAndCall", ["address", "address", "uint256"], "bool", False, True, []),
        ERC(
            "transferFromAndCall",
            ["address", "address", "uint256", "bytes"],
            "bool",
            False,
            True,
            [],
        ),
        ERC("approveAndCall", ["address", "uint256"], "bool", False, True, []),
        ERC("approveAndCall", ["address", "uint256", "bytes"], "bool", False, True, []),
    ]
    + ERC20
    + ERC165
)

ERC1363_signatures = erc_to_signatures(ERC1363)

# Review
# https://eips.ethereum.org/EIPS/eip-4524
# Must have ERC20 and ERC165

ERC4524_EVENTS = []
ERC4524 = (
    [
        ERC("safeTransfer", ["address", "uint256"], "bool", False, True, []),
        ERC("safeTransfer", ["address", "uint256", "bytes"], "bool", False, True, []),
        ERC("safeTransferFrom", ["address", "address", "uint256"], "bool", False, True, []),
        ERC(
            "safeTransferFrom", ["address", "address", "uint256", "bytes"], "bool", False, True, []
        ),
    ]
    + ERC20
    + ERC165
)

ERC4524_signatures = erc_to_signatures(ERC4524)

# Final
# https://eips.ethereum.org/EIPS/eip-4626
# Must have ERC20

ERC4626_deposit_event = ERC_EVENT(
    "Deposit",
    ["address", "address", "uint256", "uint256"],
    [True, True, False, False],
)

ERC4626_withdraw_event = ERC_EVENT(
    "Withdraw",
    ["address", "address", "address", "uint256", "uint256"],
    [True, True, True, False, False],
)

ERC4626_EVENTS = [
    ERC4626_deposit_event,
    ERC4626_withdraw_event,
]

ERC4626 = [
    ERC("asset", [], "address", True, True, []),
    ERC("totalAssets", [], "uint256", True, True, []),
    ERC("convertToShares", ["uint256"], "uint256", True, True, []),
    ERC("convertToAssets", ["uint256"], "uint256", True, True, []),
    ERC("maxDeposit", ["address"], "uint256", True, True, []),
    ERC("previewDeposit", ["uint256"], "uint256", True, True, []),
    ERC("deposit", ["uint256", "address"], "uint256", False, True, [ERC4626_deposit_event]),
    ERC("maxMint", ["address"], "uint256", True, True, []),
    ERC("previewMint", ["uint256"], "uint256", True, True, []),
    ERC("mint", ["uint256", "address"], "uint256", False, True, [ERC4626_deposit_event]),
    ERC("maxWithdraw", ["address"], "uint256", True, True, []),
    ERC("previewWithdraw", ["uint256"], "uint256", True, True, []),
    ERC(
        "withdraw",
        ["uint256", "address", "address"],
        "uint256",
        False,
        True,
        [ERC4626_withdraw_event],
    ),
    ERC("maxRedeem", ["address"], "uint256", True, True, []),
    ERC("previewRedeem", ["uint256"], "uint256", True, True, []),
    ERC(
        "redeem",
        ["uint256", "address", "address"],
        "uint256",
        False,
        True,
        [ERC4626_withdraw_event],
    ),
] + ERC20

ERC4626_signatures = erc_to_signatures(ERC4626)

ERCS = {
    "ERC20": (ERC20, ERC20_EVENTS),
    "ERC223": (ERC223, ERC223_EVENTS),
    "ERC165": (ERC165, ERC165_EVENTS),
    "ERC721": (ERC721, ERC721_EVENTS),
    "ERC1820": (ERC1820, ERC1820_EVENTS),
    "ERC777": (ERC777, ERC777_EVENTS),
    "ERC1155": (ERC1155, ERC1155_EVENTS),
    "ERC2612": (ERC2612, ERC2612_EVENTS),
    "ERC1363": (ERC1363, ERC1363_EVENTS),
    "ERC4524": (ERC4524, ERC4524_EVENTS),
    "ERC4626": (ERC4626, ERC4626_EVENTS),
}



from .array_type import ArrayType
from .elementary_type import ElementaryType
from .function_type import FunctionType
from .mapping_type import MappingType
from .user_defined_type import UserDefinedType
from .type import Type
from .type_information import TypeInformation
from .type_alias import TypeAlias, TypeAliasTopLevel, TypeAliasContract



from typing import Union, TYPE_CHECKING

from slither.core.cfg.node import NodeType, link_nodes, Node
from slither.core.cfg.scope import Scope
from slither.core.declarations.function import (
    Function,
    FunctionType,
)
from slither.core.declarations.function import ModifierStatements
from slither.core.declarations.modifier import Modifier
from slither.core.source_mapping.source_mapping import Source
from slither.core.variables.local_variable import LocalVariable
from slither.vyper_parsing.cfg.node import NodeVyper
from slither.solc_parsing.exceptions import ParsingError
from slither.vyper_parsing.variables.local_variable import LocalVariableVyper
from slither.vyper_parsing.ast.types import (
    Int,
    Call,
    Attribute,
    Name,
    Tuple as TupleVyper,
    ASTNode,
    AnnAssign,
    FunctionDef,
    Return,
    Assert,
    Compare,
    Log,
    Subscript,
    If,
    Pass,
    Assign,
    AugAssign,
    Raise,
    Expr,
    For,
    Index,
    Arg,
    Arguments,
    Continue,
    Break,
)

if TYPE_CHECKING:
    from slither.core.compilation_unit import SlitherCompilationUnit
    from slither.vyper_parsing.declarations.contract import ContractVyper


def link_underlying_nodes(node1: NodeVyper, node2: NodeVyper):
    link_nodes(node1.underlying_node, node2.underlying_node)


class FunctionVyper:
    def __init__(
        self,
        function: Function,
        function_data: FunctionDef,
        contract_parser: "ContractVyper",
    ) -> None:
        self._function = function
        self._function.name = function_data.name
        self._function.id = function_data.node_id
        self._functionNotParsed = function_data
        self._decoratorNotParsed = None
        self._local_variables_parser: list[LocalVariableVyper] = []
        self._variables_renamed = []
        self._contract_parser = contract_parser
        self._node_to_NodeVyper: dict[Node, NodeVyper] = {}

        for decorator in function_data.decorators:
            if isinstance(decorator, Call):
                # TODO handle multiple
                self._decoratorNotParsed = decorator
            elif isinstance(decorator, Name):
                if decorator.id in ["external", "public", "internal"]:
                    self._function.visibility = decorator.id
                elif decorator.id == "view":
                    self._function.view = True
                elif decorator.id == "pure":
                    self._function.pure = True
                elif decorator.id == "payable":
                    self._function.payable = True
                elif decorator.id == "nonpayable":
                    self._function.payable = False
            else:
                raise ValueError(f"Unknown decorator {decorator.id}")

        # Interfaces do not have decorators and are external
        if self._function._visibility is None:
            self._function.visibility = "external"

        self._params_was_analyzed = False
        self._content_was_analyzed = False
        self._counter_scope_local_variables = 0

        if function_data.doc_string is not None:
            function.has_documentation = True

        self._analyze_function_type()

    @property
    def underlying_function(self) -> Function:
        return self._function

    @property
    def compilation_unit(self) -> "SlitherCompilationUnit":
        return self._function.compilation_unit

    ###################################################################################
    ###################################################################################
    # region Variables
    ###################################################################################
    ###################################################################################

    @property
    def variables_renamed(
        self,
    ) -> dict[int, LocalVariableVyper]:
        return self._variables_renamed

    def _add_local_variable(self, local_var_parser: LocalVariableVyper) -> None:
        # Ensure variables name are unique for SSA conversion
        # This should not apply to actual Vyper variables currently
        # but is necessary if we have nested loops where we've created artificial variables e.g. counter_var
        if local_var_parser.underlying_variable.name:
            known_variables = [v.name for v in self._function.variables]
            while local_var_parser.underlying_variable.name in known_variables:
                local_var_parser.underlying_variable.name += (
                    f"_scope_{self._counter_scope_local_variables}"
                )
                self._counter_scope_local_variables += 1
                known_variables = [v.name for v in self._function.variables]
        # TODO no reference ID
        # if local_var_parser.reference_id is not None:
        #     self._variables_renamed[local_var_parser.reference_id] = local_var_parser
        self._function.variables_as_dict[local_var_parser.underlying_variable.name] = (
            local_var_parser.underlying_variable
        )
        self._local_variables_parser.append(local_var_parser)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Analyses
    ###################################################################################
    ###################################################################################

    @property
    def function_not_parsed(self) -> dict:
        return self._functionNotParsed

    def _analyze_function_type(self) -> None:
        if self._function.name == "__init__":
            self._function.function_type = FunctionType.CONSTRUCTOR
        elif self._function.name == "__default__":
            self._function.function_type = FunctionType.FALLBACK
        else:
            self._function.function_type = FunctionType.NORMAL

    def analyze_params(self) -> None:
        if self._params_was_analyzed:
            return

        self._params_was_analyzed = True

        params = self._functionNotParsed.args
        returns = self._functionNotParsed.returns

        if params:
            self._parse_params(params)
        if returns:
            self._parse_returns(returns)

    def analyze_content(self) -> None:
        if self._content_was_analyzed:
            return

        self._content_was_analyzed = True

        body = self._functionNotParsed.body

        if body and not isinstance(body[0], Pass):
            self._function.is_implemented = True
            self._function.is_empty = False
            self._parse_cfg(body)
            self._update_reachability(self._function.entry_point)
        else:
            self._function.is_implemented = False
            self._function.is_empty = True

        for local_var_parser in self._local_variables_parser:
            local_var_parser.analyze(self._function)

        for node_parser in self._node_to_NodeVyper.values():
            node_parser.analyze_expressions(self._function)

        self._analyze_decorator()

    def _analyze_decorator(self) -> None:
        if not self._decoratorNotParsed:
            return

        decorator = self._decoratorNotParsed
        if decorator.args:
            name = f"{decorator.func.id}({decorator.args[0].value})"
        else:
            name = decorator.func.id

        contract = self._contract_parser.underlying_contract
        compilation_unit = self._contract_parser.underlying_contract.compilation_unit
        modifier = Modifier(compilation_unit)
        modifier.name = name
        modifier.set_offset(decorator.src, compilation_unit)
        modifier.set_contract(contract)
        modifier.set_contract_declarer(contract)
        latest_entry_point = self._function.entry_point
        self._function.add_modifier(
            ModifierStatements(
                modifier=modifier,
                entry_point=latest_entry_point,
                nodes=[latest_entry_point],
            )
        )

    # endregion
    ###################################################################################
    ###################################################################################
    # region Nodes
    ###################################################################################
    ###################################################################################

    def _new_node(
        self, node_type: NodeType, src: str | Source, scope: Union[Scope, "Function"]
    ) -> NodeVyper:
        node = self._function.new_node(node_type, src, scope)
        node_parser = NodeVyper(node)
        self._node_to_NodeVyper[node] = node_parser
        return node_parser

    # endregion
    ###################################################################################
    ###################################################################################
    # region Parsing function
    ###################################################################################
    ###################################################################################
    def _update_reachability(self, node: Node) -> None:
        if node.is_reachable:
            return
        node.set_is_reachable(True)
        for son in node.sons:
            self._update_reachability(son)

    def _parse_cfg(self, cfg: list[ASTNode]) -> None:
        entry_node = self._new_node(NodeType.ENTRYPOINT, "-1:-1:-1", self.underlying_function)
        self._function.entry_point = entry_node.underlying_node
        scope = Scope(True, False, self.underlying_function)

        def parse_statement(
            curr_node: NodeVyper,
            expr: ASTNode,
            continue_destination=None,
            break_destination=None,
        ) -> NodeVyper:
            if isinstance(expr, AnnAssign):
                local_var = LocalVariable()
                local_var.set_function(self._function)
                local_var.set_offset(expr.src, self._function.compilation_unit)

                local_var_parser = LocalVariableVyper(local_var, expr)
                self._add_local_variable(local_var_parser)

                new_node = self._new_node(NodeType.VARIABLE, expr.src, scope)
                if expr.value is not None:
                    local_var.initialized = True
                    new_node.add_unparsed_expression(expr.value)
                new_node.underlying_node.add_variable_declaration(local_var)
                link_underlying_nodes(curr_node, new_node)

                curr_node = new_node

            elif isinstance(expr, (AugAssign, Assign)):
                new_node = self._new_node(NodeType.EXPRESSION, expr.src, scope)
                new_node.add_unparsed_expression(expr)
                link_underlying_nodes(curr_node, new_node)

                curr_node = new_node

            elif isinstance(expr, Expr):
                # TODO This is a workaround to handle Vyper putting payable/view in the function body... https://github.com/vyperlang/vyper/issues/3578
                if not isinstance(expr.value, Name):
                    new_node = self._new_node(NodeType.EXPRESSION, expr.src, scope)
                    new_node.add_unparsed_expression(expr.value)
                    link_underlying_nodes(curr_node, new_node)

                    curr_node = new_node

            elif isinstance(expr, For):
                node_startLoop = self._new_node(NodeType.STARTLOOP, expr.src, scope)
                node_endLoop = self._new_node(NodeType.ENDLOOP, expr.src, scope)

                link_underlying_nodes(curr_node, node_startLoop)

                local_var = LocalVariable()
                local_var.set_function(self._function)
                local_var.set_offset(expr.src, self._function.compilation_unit)

                counter_var = AnnAssign(
                    expr.target.src,
                    expr.target.node_id,
                    target=Name("-1:-1:-1", -1, "counter_var"),
                    annotation=Name("-1:-1:-1", -1, "uint256"),
                    value=Int("-1:-1:-1", -1, 0),
                )
                local_var_parser = LocalVariableVyper(local_var, counter_var)
                self._add_local_variable(local_var_parser)
                counter_node = self._new_node(NodeType.VARIABLE, expr.src, scope)
                local_var.initialized = True
                counter_node.add_unparsed_expression(counter_var.value)
                counter_node.underlying_node.add_variable_declaration(local_var)

                link_underlying_nodes(node_startLoop, counter_node)

                node_condition = None
                if isinstance(expr.iter, (Attribute, Name)):
                    # HACK
                    # The loop variable is not annotated so we infer its type by looking at the type of the iterator
                    if isinstance(expr.iter, Attribute):  # state variable
                        iter_expr = expr.iter
                        loop_iterator = list(
                            filter(
                                lambda x: x._variable.name == iter_expr.attr,
                                self._contract_parser._variables_parser,
                            )
                        )[0]

                    else:  # local variable
                        iter_expr = expr.iter
                        loop_iterator = list(
                            filter(
                                lambda x: x._variable.name == iter_expr.id,
                                self._local_variables_parser,
                            )
                        )[0]

                    # TODO use expr.src instead of -1:-1:1?
                    cond_expr = Compare(
                        "-1:-1:-1",
                        -1,
                        left=Name("-1:-1:-1", -1, "counter_var"),
                        op="<=",
                        right=Call(
                            "-1:-1:-1",
                            -1,
                            func=Name("-1:-1:-1", -1, "len"),
                            args=[iter_expr],
                            keywords=[],
                            keyword=None,
                        ),
                    )
                    node_condition = self._new_node(NodeType.IFLOOP, expr.src, scope)
                    node_condition.add_unparsed_expression(cond_expr)

                    if loop_iterator._elem_to_parse.value.id == "DynArray":
                        loop_var_annotation = loop_iterator._elem_to_parse.slice.value.elements[0]
                    else:
                        loop_var_annotation = loop_iterator._elem_to_parse.value

                    value = Subscript(
                        "-1:-1:-1",
                        -1,
                        value=Name("-1:-1:-1", -1, loop_iterator._variable.name),
                        slice=Index("-1:-1:-1", -1, value=Name("-1:-1:-1", -1, "counter_var")),
                    )
                    loop_var = AnnAssign(
                        expr.target.src,
                        expr.target.node_id,
                        target=expr.target,
                        annotation=loop_var_annotation,
                        value=value,
                    )

                elif isinstance(expr.iter, Call):  # range
                    range_val = expr.iter.args[0]
                    cond_expr = Compare(
                        "-1:-1:-1",
                        -1,
                        left=Name("-1:-1:-1", -1, "counter_var"),
                        op="<=",
                        right=range_val,
                    )
                    node_condition = self._new_node(NodeType.IFLOOP, expr.src, scope)
                    node_condition.add_unparsed_expression(cond_expr)
                    loop_var = AnnAssign(
                        expr.target.src,
                        expr.target.node_id,
                        target=expr.target,
                        annotation=Name("-1:-1:-1", -1, "uint256"),
                        value=Name("-1:-1:-1", -1, "counter_var"),
                    )

                else:
                    raise NotImplementedError

                # After creating condition node, we link it declaration of the loop variable
                link_underlying_nodes(counter_node, node_condition)

                # Create an expression for the loop increment (counter_var += 1)
                loop_increment = AugAssign(
                    "-1:-1:-1",
                    -1,
                    target=Name("-1:-1:-1", -1, "counter_var"),
                    op="+=",
                    value=Int("-1:-1:-1", -1, 1),
                )
                node_increment = self._new_node(NodeType.EXPRESSION, expr.src, scope)
                node_increment.add_unparsed_expression(loop_increment)
                link_underlying_nodes(node_increment, node_condition)

                continue_destination = node_increment
                break_destination = node_endLoop

                # We assign the index variable or range variable in the loop body on each iteration
                expr.body.insert(0, loop_var)
                body_node = None
                new_node = node_condition
                for stmt in expr.body:
                    body_node = parse_statement(
                        new_node, stmt, continue_destination, break_destination
                    )
                    new_node = body_node

                if body_node is not None:
                    link_underlying_nodes(body_node, node_increment)

                link_underlying_nodes(node_condition, node_endLoop)

                curr_node = node_endLoop

            elif isinstance(expr, Continue):
                new_node = self._new_node(NodeType.CONTINUE, expr.src, scope)
                link_underlying_nodes(curr_node, new_node)
                link_underlying_nodes(new_node, continue_destination)

            elif isinstance(expr, Break):
                new_node = self._new_node(NodeType.BREAK, expr.src, scope)
                link_underlying_nodes(curr_node, new_node)
                link_underlying_nodes(new_node, break_destination)

            elif isinstance(expr, Return):
                new_node = self._new_node(NodeType.RETURN, expr.src, scope)
                if expr.value is not None:
                    new_node.add_unparsed_expression(expr.value)

                link_underlying_nodes(curr_node, new_node)
                curr_node = new_node

            elif isinstance(expr, Assert):
                new_node = self._new_node(NodeType.EXPRESSION, expr.src, scope)
                new_node.add_unparsed_expression(expr)

                link_underlying_nodes(curr_node, new_node)
                curr_node = new_node

            elif isinstance(expr, Log):
                new_node = self._new_node(NodeType.EXPRESSION, expr.src, scope)
                new_node.add_unparsed_expression(expr.value)

                link_underlying_nodes(curr_node, new_node)
                curr_node = new_node

            elif isinstance(expr, If):
                condition_node = self._new_node(NodeType.IF, expr.test.src, scope)
                condition_node.add_unparsed_expression(expr.test)

                endIf_node = self._new_node(NodeType.ENDIF, expr.src, scope)

                true_node = None
                new_node = condition_node
                for stmt in expr.body:
                    true_node = parse_statement(
                        new_node, stmt, continue_destination, break_destination
                    )
                    new_node = true_node

                link_underlying_nodes(true_node, endIf_node)

                false_node = None
                new_node = condition_node
                for stmt in expr.orelse:
                    false_node = parse_statement(
                        new_node, stmt, continue_destination, break_destination
                    )
                    new_node = false_node

                if false_node is not None:
                    link_underlying_nodes(false_node, endIf_node)

                else:
                    link_underlying_nodes(condition_node, endIf_node)

                link_underlying_nodes(curr_node, condition_node)
                curr_node = endIf_node

            elif isinstance(expr, Pass):
                pass
            elif isinstance(expr, Raise):
                new_node = self._new_node(NodeType.EXPRESSION, expr.src, scope)
                new_node.add_unparsed_expression(expr)
                link_underlying_nodes(curr_node, new_node)
                curr_node = new_node

            else:
                raise ParsingError(f"Statement not parsed {expr.__class__.__name__} {expr}")

            return curr_node

        curr_node = entry_node
        for expr in cfg:
            curr_node = parse_statement(curr_node, expr)

    # endregion
    ###################################################################################
    ###################################################################################

    def _add_param(self, param: Arg, initialized: bool = False) -> LocalVariableVyper:
        local_var = LocalVariable()
        local_var.set_function(self._function)
        local_var.set_offset(param.src, self._function.compilation_unit)
        local_var_parser = LocalVariableVyper(local_var, param)

        if initialized:
            local_var.initialized = True

        if local_var.location == "default":
            local_var.set_location("memory")

        self._add_local_variable(local_var_parser)
        return local_var_parser

    def _parse_params(self, params: Arguments):
        self._function.parameters_src().set_offset(params.src, self._function.compilation_unit)
        if params.defaults:
            self._function._default_args_as_expressions = params.defaults
        for param in params.args:
            local_var = self._add_param(param)
            self._function.add_parameters(local_var.underlying_variable)

    def _parse_returns(self, returns: Name | TupleVyper | Subscript):
        self._function.returns_src().set_offset(returns.src, self._function.compilation_unit)
        # Only the type of the arg is given, not a name. We create an `Arg` with an empty name
        # so that the function has the correct return type in its signature but doesn't clash with
        # other identifiers during name resolution (`find_variable`).
        if isinstance(returns, (Name, Subscript)):
            local_var = self._add_param(Arg(returns.src, returns.node_id, "", annotation=returns))
            self._function.add_return(local_var.underlying_variable)
        else:
            assert isinstance(returns, TupleVyper)
            for ret in returns.elements:
                local_var = self._add_param(Arg(ret.src, ret.node_id, "", annotation=ret))
                self._function.add_return(local_var.underlying_variable)

    ###################################################################################
    ###################################################################################



"""
Module detecting shadowing of state variables
"""

from slither.core.declarations import Contract
from slither.core.variables.state_variable import StateVariable
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.detectors.shadowing.common import is_upgradable_gap_variable
from slither.utils.output import Output


def detect_shadowing(contract: Contract) -> list[list[StateVariable]]:
    ret = []
    variables_fathers = []
    for father in contract.inheritance:
        if any(f.is_implemented for f in father.functions + father.modifiers):
            variables_fathers += father.state_variables_declared

    for var in contract.state_variables_declared:
        # Ignore __gap variables for updatable contracts
        if is_upgradable_gap_variable(contract, var):
            continue

        shadow = [v for v in variables_fathers if v.name == var.name]
        if shadow:
            ret.append([var] + shadow)
    return ret


class StateShadowing(AbstractDetector):
    """
    Shadowing of state variable
    """

    ARGUMENT = "shadowing-state"
    HELP = "State variables shadowing"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#state-variable-shadowing"

    WIKI_TITLE = "State variable shadowing"
    WIKI_DESCRIPTION = "Detection of state variables shadowed."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract BaseContract{
    address owner;

    modifier isOwner(){
        require(owner == msg.sender);
        _;
    }

}

contract DerivedContract is BaseContract{
    address owner;

    constructor(){
        owner = msg.sender;
    }

    function withdraw() isOwner() external{
        msg.sender.transfer(this.balance);
    }
}
```
`owner` of `BaseContract` is never assigned and the modifier `isOwner` does not work."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Remove the state variable shadowing."

    def _detect(self) -> list[Output]:
        """Detect shadowing

        Recursively visit the calls
        Returns:
            list: {'vuln', 'filename,'contract','func', 'shadow'}

        """
        results = []
        for c in self.contracts:
            shadowing = detect_shadowing(c)
            if shadowing:
                for all_variables in shadowing:
                    shadow = all_variables[0]
                    variables = all_variables[1:]
                    info: DETECTOR_INFO = [shadow, " shadows:\n"]
                    for var in variables:
                        info += ["\t- ", var, "\n"]

                    res = self.generate_result(info)
                    results.append(res)

        return results



import abc
import logging
from pathlib import Path
from slither.core.compilation_unit import SlitherCompilationUnit
from slither.formatters.utils.patches import apply_patch, create_diff
from slither.tools.mutator.utils.testing_generated_mutant import test_patch
from slither.core.declarations import Contract
from slither.utils.colors import red

logger = logging.getLogger("Slither-Mutate")


class IncorrectMutatorInitialization(Exception):
    pass


class AbstractMutator(metaclass=abc.ABCMeta):
    NAME = ""
    HELP = ""

    def __init__(
        self,
        compilation_unit: SlitherCompilationUnit,
        timeout: int,
        testing_command: str,
        testing_directory: str,
        contract_instance: Contract,
        solc_remappings: str | None,
        verbose: bool,
        output_folder: Path,
        dont_mutate_line: list[int],
        rate: int = 10,
        seed: int | None = None,
        target_selectors: set[int] | None = None,
        target_modifiers: set[str] | None = None,
    ) -> None:
        self.compilation_unit = compilation_unit
        self.slither = compilation_unit.core
        self.seed = seed
        self.rate = rate
        self.test_command = testing_command
        self.test_directory = testing_directory
        self.timeout = timeout
        self.solc_remappings = solc_remappings
        self.verbose = verbose
        self.output_folder = output_folder
        self.contract = contract_instance
        self.in_file = self.contract.source_mapping.filename.absolute
        self.dont_mutate_line = dont_mutate_line
        self.target_selectors = target_selectors
        self.target_modifiers = target_modifiers
        # total revert/comment/tweak mutants that were generated and compiled
        self.total_mutant_counts = [0, 0, 0]
        # total caught revert/comment/tweak mutants
        self.caught_mutant_counts = [0, 0, 0]

        if not self.NAME:
            raise IncorrectMutatorInitialization(
                f"NAME is not initialized {self.__class__.__name__}"
            )

        if not self.HELP:
            raise IncorrectMutatorInitialization(
                f"HELP is not initialized {self.__class__.__name__}"
            )

        if rate < 0 or rate > 100:
            raise IncorrectMutatorInitialization(
                f"rate must be between 0 and 100 {self.__class__.__name__}"
            )

    def should_mutate_node(self, node) -> bool:
        return (
            node.source_mapping.lines[0] not in self.dont_mutate_line
            and node.source_mapping.filename.absolute == self.in_file
        )

    def should_mutate_function(self, function) -> bool:
        """Check if function/modifier should be mutated based on target_selectors"""
        if self.target_selectors is None:
            return True  # No filter, mutate all

        from slither.utils.function import get_function_id
        from slither.core.declarations import Modifier

        if isinstance(function, Modifier):
            # For modifiers, check if in target_modifiers set
            return bool(self.target_modifiers and function.name in self.target_modifiers)

        # For functions, check selector
        try:
            func_selector = get_function_id(function.solidity_signature)
            return func_selector in self.target_selectors
        except Exception:  # pylint: disable=broad-except
            return False

    @abc.abstractmethod
    def _mutate(self) -> dict:
        """Abstract placeholder, will be overwritten by each mutator"""
        return {}

    def mutate(self) -> tuple[list[int], list[int], list[int]]:
        all_patches: dict = {}

        # call _mutate implementation of different mutation types
        try:
            (all_patches) = self._mutate()
        except Exception as e:
            logger.error(red("%s mutator failed in %s: %s"), self.NAME, self.contract.name, str(e))

        if "patches" not in all_patches:
            logger.debug("No patches found by %s", self.NAME)
            return [0, 0, 0], [0, 0, 0], self.dont_mutate_line

        for file in all_patches["patches"]:  # Note: This should only loop over a single file
            original_txt = self.slither.source_code[file].encode("utf8")
            patches = all_patches["patches"][file]
            patches.sort(key=lambda x: x["start"])
            for patch in patches:
                # test the patch
                patch_was_caught = test_patch(
                    self.output_folder,
                    file,
                    patch,
                    self.test_command,
                    self.NAME,
                    self.timeout,
                    self.solc_remappings,
                    self.verbose,
                )

                # skip mutants that couldn't compile
                if patch_was_caught == 2:
                    continue

                # Add all mutants that could compile to the totals by type
                if self.NAME == "RR":
                    self.total_mutant_counts[0] += 1
                elif self.NAME == "CR":
                    self.total_mutant_counts[1] += 1
                else:
                    self.total_mutant_counts[2] += 1

                # count caught mutants
                if patch_was_caught == 1:
                    if self.NAME == "RR":
                        self.caught_mutant_counts[0] += 1
                    elif self.NAME == "CR":
                        self.caught_mutant_counts[1] += 1
                    else:
                        self.caught_mutant_counts[2] += 1

                # record uncaught mutants to skip known-under-tested lines
                if patch_was_caught == 0:
                    if self.NAME == "RR":
                        self.dont_mutate_line.append(patch["line_number"])
                    elif self.NAME == "CR":
                        self.dont_mutate_line.append(patch["line_number"])

                    patched_txt, _ = apply_patch(original_txt, patch, 0)
                    diff = create_diff(self.compilation_unit, original_txt, patched_txt, file)
                    if not diff:
                        logger.info(f"Impossible to generate patch; empty {patches}")

                    # add uncaught mutant patches to a output file
                    with (self.output_folder / "patches_files.txt").open(
                        "a", encoding="utf8"
                    ) as patches_file:
                        patches_file.write(diff + "\n")

        return self.total_mutant_counts, self.caught_mutant_counts, self.dont_mutate_line



import traceback
import logging
from pathlib import Path
import hashlib

logger = logging.getLogger("Slither-Mutate")

HashedPath = str
backuped_files: dict[str, HashedPath] = {}


def backup_source_file(source_code: dict, output_folder: Path) -> dict[str, HashedPath]:
    """
    function to backup the source file
    returns: dictionary of duplicated files
    """
    output_folder.mkdir(exist_ok=True, parents=True)
    for file_path, content in source_code.items():
        path_hash = hashlib.md5(bytes(file_path, "utf8")).hexdigest()
        (output_folder / path_hash).write_text(content, encoding="utf8")

        backuped_files[file_path] = (output_folder / path_hash).as_posix()

    return backuped_files


def transfer_and_delete(files_dict: dict[str, HashedPath]) -> None:
    """function to transfer the original content to the sol file after campaign"""
    try:
        files_dict_copy = files_dict.copy()
        for original_path, hashed_path in files_dict_copy.items():
            with open(hashed_path, encoding="utf8") as duplicated_file:
                content = duplicated_file.read()

            with open(original_path, "w", encoding="utf8") as original_file:
                original_file.write(content)

            Path(hashed_path).unlink()

            # delete elements from the global dict
            del backuped_files[original_path]

    except FileNotFoundError as e:
        logger.error("Error transferring content: %s", e)


global_counter = {}


def create_mutant_file(output_folder: Path, file: str, rule: str) -> None:
    """function to create new mutant file"""
    try:
        if rule not in global_counter:
            global_counter[rule] = 0

        file_path = Path(file)
        # Read content from the duplicated file
        content = file_path.read_text(encoding="utf8")

        # Write content to the original file
        mutant_name = file_path.stem
        # create folder for each contract
        mutation_dir = output_folder / mutant_name
        mutation_dir.mkdir(parents=True, exist_ok=True)

        mutation_filename = f"{mutant_name}_{rule}_{global_counter[rule]}.sol"
        with (mutation_dir / mutation_filename).open("w", encoding="utf8") as mutant_file:
            mutant_file.write(content)
        global_counter[rule] += 1

        # reset the file
        duplicate_content = Path(backuped_files[file]).read_text("utf8")

        with open(file, "w", encoding="utf8") as source_file:
            source_file.write(duplicate_content)

    except Exception as e:
        logger.error(f"Error creating mutant: {e}")
        traceback_str = traceback.format_exc()
        logger.error(traceback_str)  # Log the stack trace


def reset_file(file: str) -> None:
    """function to reset the file"""
    try:
        # reset the file
        with open(backuped_files[file], encoding="utf8") as duplicated_file:
            duplicate_content = duplicated_file.read()

        with open(file, "w", encoding="utf8") as source_file:
            source_file.write(duplicate_content)

    except Exception as e:
        logger.error("Error resetting file: %s", e)


def get_sol_file_list(codebase: Path, ignore_paths: list[str] | None) -> list[str]:
    """
    function to get the contracts list
    returns: list of .sol files
    """
    sol_file_list = []
    if ignore_paths is None:
        ignore_paths = []

    # if input is contract file
    if codebase.is_file():
        return [codebase.as_posix()]

    # if input is folder
    if codebase.is_dir():
        for file_name in codebase.rglob("*.sol"):
            if file_name.is_file() and not any(part in ignore_paths for part in file_name.parts):
                sol_file_list.append(file_name.as_posix())

    return sol_file_list



from pathlib import Path

from slither import Slither
from slither.slithir.operations.unpack import Unpack

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data" / "tuple_assign"


def test_tuple_order(solc_binary_path) -> None:
    """Test that the correct tuple components are projected when some components are left empty."""
    solc_path = solc_binary_path("0.8.15")
    slither = Slither(Path(TEST_DATA_DIR, "test_tuple_assign.sol").as_posix(), solc=solc_path)
    fn = slither.get_contract_from_name("TestEmptyComponent")[0].get_function_from_canonical_name(
        "TestEmptyComponent.test()"
    )
    unpacks = [ir for ir in fn.slithir_operations if isinstance(ir, Unpack)]
    assert unpacks[0].index == 0
    assert unpacks[1].index == 2
    assert unpacks[2].index == 3


def test_tuple_reassign(solc_binary_path) -> None:
    """Test that tuple component indexes are not reused across assignments"""
    solc_path = solc_binary_path("0.8.15")
    slither = Slither(Path(TEST_DATA_DIR, "test_tuple_assign.sol").as_posix(), solc=solc_path)
    fn = slither.get_contract_from_name("TestTupleReassign")[0].get_function_from_canonical_name(
        "TestTupleReassign.test()"
    )
    unpacks = [ir for ir in fn.slithir_operations if isinstance(ir, Unpack)]
    assert len(unpacks) == 5
    assert unpacks[3].index == 0
    assert unpacks[4].index == 1



from pathlib import Path

from slither import Slither


TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"
CUSTOM_COMMENTS_TEST_DATA_DIR = Path(TEST_DATA_DIR, "custom_comments")


def test_upgradeable_comments(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.10")
    slither = Slither(Path(CUSTOM_COMMENTS_TEST_DATA_DIR, "upgrade.sol").as_posix(), solc=solc_path)
    compilation_unit = slither.compilation_units[0]
    proxy = compilation_unit.get_contract_from_name("Proxy")[0]

    assert proxy.is_upgradeable_proxy

    v0 = compilation_unit.get_contract_from_name("V0")[0]

    assert v0.is_upgradeable
    print(v0.upgradeable_version)
    assert v0.upgradeable_version == "version-0"

    v1 = compilation_unit.get_contract_from_name("V1")[0]
    assert v0.is_upgradeable
    assert v1.upgradeable_version == "version_1"


def test_contract_comments(solc_binary_path) -> None:
    comments = " @title Test Contract\n @dev Test comment"

    solc_path = solc_binary_path("0.8.10")
    slither = Slither(
        Path(CUSTOM_COMMENTS_TEST_DATA_DIR, "contract_comment.sol").as_posix(), solc=solc_path
    )
    compilation_unit = slither.compilation_units[0]
    contract = compilation_unit.get_contract_from_name("A")[0]

    assert contract.comments == comments

    # Old solc versions have a different parsing of comments
    # the initial space (after *) is also not kept on every line
    comments = "@title Test Contract\n@dev Test comment"
    solc_path = solc_binary_path("0.5.16")
    slither = Slither(
        Path(CUSTOM_COMMENTS_TEST_DATA_DIR, "contract_comment.sol").as_posix(), solc=solc_path
    )
    compilation_unit = slither.compilation_units[0]
    contract = compilation_unit.get_contract_from_name("A")[0]

    assert contract.comments == comments

    # Test with legacy AST
    comments = "@title Test Contract\n@dev Test comment"
    slither = Slither(
        Path(CUSTOM_COMMENTS_TEST_DATA_DIR, "contract_comment.sol").as_posix(),
        solc_force_legacy_json=True,
        solc=solc_path,
    )
    compilation_unit = slither.compilation_units[0]
    contract = compilation_unit.get_contract_from_name("A")[0]

    assert contract.comments == comments



from slither.slithir.operations.lvalue import OperationWithLValue
from slither.slithir.variables.temporary import TemporaryVariable


class TmpNewContract(OperationWithLValue):
    def __init__(self, contract_name: str, lvalue: TemporaryVariable) -> None:
        super().__init__()
        self._contract_name = contract_name
        self._lvalue = lvalue
        self._call_value = None
        self._call_salt = None

    @property
    def contract_name(self) -> str:
        return self._contract_name

    @property
    def call_value(self):
        return self._call_value

    @call_value.setter
    def call_value(self, v):
        self._call_value = v

    @property
    def call_salt(self):
        return self._call_salt

    @call_salt.setter
    def call_salt(self, s):
        self._call_salt = s

    @property
    def read(self):
        return []

    def __str__(self):
        return f"{self.lvalue} = new {self.contract_name}"



from typing import TYPE_CHECKING, Optional

from slither.core.declarations import Event, Enum, Structure
from slither.core.declarations.contract import Contract
from slither.core.declarations.custom_error import CustomError
from slither.core.declarations.function import Function
from slither.core.declarations.function_contract import FunctionContract
from slither.core.declarations.solidity_variables import (
    SOLIDITY_FUNCTIONS,
    SOLIDITY_VARIABLES,
    SolidityFunction,
    SolidityVariable,
)
from slither.core.variables.variable import Variable
from slither.solc_parsing.exceptions import VariableNotFound

if TYPE_CHECKING:
    from slither.vyper_parsing.declarations.function import FunctionVyper


def _find_variable_in_function_parser(
    var_name: str,
    function_parser: Optional["FunctionVyper"],
) -> Variable | None:
    if function_parser is None:
        return None
    func_variables = function_parser.variables_as_dict
    if var_name in func_variables:
        return func_variables[var_name]

    return None


def _find_in_contract(
    var_name: str,
    contract: Contract | None,
    contract_declarer: Contract | None,
) -> Variable | Function | Contract | Event | Enum | Structure | CustomError | None:
    if contract is None or contract_declarer is None:
        return None

    # variable are looked from the contract declarer
    contract_variables = contract.variables_as_dict
    if var_name in contract_variables:
        return contract_variables[var_name]

    functions = {f.name: f for f in contract.functions if not f.is_shadowed}
    if var_name in functions:
        return functions[var_name]

    # structures are looked on the contract declarer
    structures = contract.structures_as_dict
    if var_name in structures:
        return structures[var_name]

    events = contract.events_as_dict
    if var_name in events:
        return events[var_name]

    enums = contract.enums_as_dict
    if var_name in enums:
        return enums[var_name]

    # If the enum is referred as its name rather than its canonicalName
    enums = {e.name: e for e in contract.enums}
    if var_name in enums:
        return enums[var_name]

    return None


def find_variable(
    var_name: str,
    caller_context: FunctionContract | Contract,
    is_self: bool = False,
) -> tuple[
    Variable | Function | Contract | SolidityVariable | SolidityFunction | Event | Enum | Structure
]:
    """
    Return the variable found and a boolean indicating if the variable was created
    If the variable was created, it has no source mapping, and it the caller must add it

    :param var_name:
    :type var_name:
    :param caller_context:
    :type caller_context:
    :param is_self:
    :type is_self:
    :return:
    :rtype:
    """

    from slither.vyper_parsing.declarations.function import (
        FunctionVyper,
    )

    if isinstance(caller_context, Contract):
        current_scope = caller_context.file_scope
        next_context = caller_context
    else:
        current_scope = caller_context.contract.file_scope
        next_context = caller_context.contract

    function_parser: FunctionVyper | None = (
        caller_context if isinstance(caller_context, FunctionContract) else None
    )

    # If a local shadows a state variable but the attribute is `self`, we want to
    # return the state variable and not the local.
    if not is_self:
        ret1 = _find_variable_in_function_parser(var_name, function_parser)
        if ret1:
            return ret1

    ret = _find_in_contract(var_name, next_context, caller_context)
    if ret:
        return ret

    if var_name in current_scope.variables:
        return current_scope.variables[var_name]

    # Could refer to any enum
    all_enumss = [c.enums_as_dict for c in current_scope.contracts.values()]
    all_enums = {k: v for d in all_enumss for k, v in d.items()}
    if var_name in all_enums:
        return all_enums[var_name]

    contracts = current_scope.contracts
    if var_name in contracts:
        return contracts[var_name]

    if var_name in SOLIDITY_VARIABLES:
        return SolidityVariable(var_name)

    if f"{var_name}()" in SOLIDITY_FUNCTIONS:
        return SolidityFunction(f"{var_name}()")

    if f"{var_name}()" in next_context.events_as_dict:
        return next_context.events_as_dict[f"{var_name}()"]

    raise VariableNotFound(f"Variable not found: {var_name} (context {caller_context})")



"""
Module detecting misuse of Boolean constants
"""

from slither.core.cfg.node import Node
from slither.core.declarations import Function
from slither.core.declarations.contract import Contract
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations import (
    Binary,
    BinaryType,
)
from slither.slithir.variables import Constant
from slither.utils.output import Output


class BooleanEquality(AbstractDetector):
    """
    Boolean constant equality
    """

    ARGUMENT = "boolean-equal"
    HELP = "Comparison to boolean constant"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#boolean-equality"

    WIKI_TITLE = "Boolean equality"
    WIKI_DESCRIPTION = """Detects the comparison to boolean constants."""

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract A {
	function f(bool x) public {
		// ...
        if (x == true) { // bad!
           // ...
        }
		// ...
	}
}
```
Boolean constants can be used directly and do not need to be compare to `true` or `false`."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = """Remove the equality to the boolean constant."""

    @staticmethod
    def _detect_boolean_equality(
        contract: Contract,
    ) -> list[tuple[Function, set[Node]]]:
        # Create our result set.
        results: list[tuple[Function, set[Node]]] = []

        # Loop for each function and modifier.

        for function in contract.functions_and_modifiers_declared:
            f_results = set()

            # Loop for every node in this function, looking for boolean constants
            for node in function.nodes:
                for ir in node.irs:
                    if isinstance(ir, Binary):
                        if ir.type in [BinaryType.EQUAL, BinaryType.NOT_EQUAL]:
                            for r in ir.read:
                                if isinstance(r, Constant):
                                    if isinstance(r.value, bool):
                                        f_results.add(node)
                results.append((function, f_results))

        # Return the resulting set of nodes with improper uses of Boolean constants
        return results

    def _detect(self) -> list[Output]:
        """
        Detect Boolean constant misuses
        """
        results = []
        for contract in self.contracts:
            boolean_constant_misuses = self._detect_boolean_equality(contract)
            for func, nodes in boolean_constant_misuses:
                for node in nodes:
                    info: DETECTOR_INFO = [
                        func,
                        " compares to a boolean constant:\n\t-",
                        node,
                        "\n",
                    ]

                    res = self.generate_result(info)
                    results.append(res)

        return results



from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


class PythDeprecatedFunctions(AbstractDetector):
    """
    Documentation: This detector finds deprecated Pyth function calls
    """

    ARGUMENT = "pyth-deprecated-functions"
    HELP = "Detect Pyth deprecated functions"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#pyth-deprecated-functions"
    WIKI_TITLE = "Pyth deprecated functions"
    WIKI_DESCRIPTION = "Detect when a Pyth deprecated function is used"
    WIKI_RECOMMENDATION = (
        "Do not use deprecated Pyth functions. Visit https://api-reference.pyth.network/."
    )

    WIKI_EXPLOIT_SCENARIO = """
```solidity
import "@pythnetwork/pyth-sdk-solidity/IPyth.sol";
import "@pythnetwork/pyth-sdk-solidity/PythStructs.sol";

contract C {

    IPyth pyth;

    constructor(IPyth _pyth) {
        pyth = _pyth;
    }

    function A(bytes32 priceId) public {
        PythStructs.Price memory price = pyth.getPrice(priceId);
        ...
    }
}
```
The function `A` uses the deprecated `getPrice` Pyth function.
"""

    def _detect(self):
        DEPRECATED_PYTH_FUNCTIONS = [
            "getValidTimePeriod",
            "getEmaPrice",
            "getPrice",
        ]
        results: list[Output] = []

        for contract in self.compilation_unit.contracts_derived:
            for target_contract, ir in contract.all_high_level_calls:
                if (
                    target_contract.name == "IPyth"
                    and ir.function_name in DEPRECATED_PYTH_FUNCTIONS
                ):
                    info: DETECTOR_INFO = [
                        "The following Pyth deprecated function is used\n\t- ",
                        ir.node,
                        "\n",
                    ]

                    res = self.generate_result(info)
                    results.append(res)

        return results



"""
Module detecting state uninitialized variables
Recursively check the called functions

The heuristic checks:
- state variables including mappings/refs
- LibraryCalls, InternalCalls, InternalDynamicCalls with storage variables

Only analyze "leaf" contracts (contracts that are not inherited by another contract)
"""

from slither.core.declarations import Function
from slither.core.declarations.contract import Contract
from slither.core.variables import Variable
from slither.core.variables.state_variable import StateVariable
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations import InternalCall, LibraryCall
from slither.slithir.variables import ReferenceVariable
from slither.utils.output import Output


class UninitializedStateVarsDetection(AbstractDetector):
    """
    Constant function detector
    """

    ARGUMENT = "uninitialized-state"
    HELP = "Uninitialized state variables"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#uninitialized-state-variables"

    WIKI_TITLE = "Uninitialized state variables"
    WIKI_DESCRIPTION = "Uninitialized state variables."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Uninitialized{
    address destination;

    function transfer() payable public{
        destination.transfer(msg.value);
    }
}
```
Bob calls `transfer`. As a result, the Ether are sent to the address `0x0` and are lost.
"""
    # endregion wiki_exploit_scenario

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """
Initialize all the variables. If a variable is meant to be initialized to zero, explicitly set it to zero to improve code readability.
"""
    # endregion wiki_recommendation

    @staticmethod
    def _written_variables(contract: Contract) -> list[StateVariable]:
        ret = []

        for f in contract.all_functions_called + contract.modifiers:
            for n in f.nodes:
                ret += n.state_variables_written
                for ir in n.irs:
                    if isinstance(ir, (LibraryCall, InternalCall)):
                        idx = 0
                        if ir.function:
                            for param in ir.function.parameters:
                                if param.location == "storage":
                                    # If its a storage variable, add either the variable
                                    # Or the variable it points to if its a reference
                                    if isinstance(ir.arguments[idx], ReferenceVariable):
                                        ret.append(ir.arguments[idx].points_to_origin)
                                    else:
                                        ret.append(ir.arguments[idx])
                                idx = idx + 1

        return ret

    def _variable_written_in_proxy(self):
        # Hack to memoize without having it define in the init
        if hasattr(self, "__variables_written_in_proxy"):
            return self.__variables_written_in_proxy

        variables_written_in_proxy = []
        for c in self.compilation_unit.contracts:
            if c.is_upgradeable_proxy:
                variables_written_in_proxy += self._written_variables(c)

        self.__variables_written_in_proxy = list({v.name for v in variables_written_in_proxy})
        return self.__variables_written_in_proxy

    def _written_variables_in_proxy(self, contract: Contract) -> list[StateVariable]:
        variables: list[StateVariable] = []
        if contract.is_upgradeable:
            variables_name_written_in_proxy = self._variable_written_in_proxy()
            if variables_name_written_in_proxy:
                variables_in_contract = [
                    contract.get_state_variable_from_name(v)
                    for v in variables_name_written_in_proxy
                ]
                variables += [v for v in variables_in_contract if v]
        return list(set(variables))

    @staticmethod
    def _read_variables(contract: Contract) -> list[StateVariable]:
        ret = []
        for f in contract.all_functions_called:
            if isinstance(f, Function):
                ret += f.state_variables_read
        for m in contract.modifiers:
            ret += m.state_variables_read
        return ret

    def _detect_uninitialized(self, contract: Contract) -> list[tuple[Variable, list[Function]]]:
        written_variables = self._written_variables(contract)
        written_variables += self._written_variables_in_proxy(contract)
        read_variables = self._read_variables(contract)
        return [
            (variable, contract.get_functions_reading_from_variable(variable))
            for variable in contract.state_variables
            if variable not in written_variables
            and not variable.expression
            and variable in read_variables
        ]

    def _detect(self) -> list[Output]:
        """Detect uninitialized state variables

        Recursively visit the calls
        Returns:
            dict: [contract name] = set(state variable uninitialized)
        """
        results = []
        for c in self.compilation_unit.contracts_derived:
            ret = self._detect_uninitialized(c)
            for variable, functions in ret:
                info: DETECTOR_INFO = [variable, " is never initialized. It is used in:\n"]

                for f in functions:
                    info += ["\t- ", f, "\n"]

                json = self.generate_result(info)
                results.append(json)

        return results



from slither.slithir.operations.lvalue import OperationWithLValue

from slither.slithir.utils.utils import is_valid_lvalue
from slither.core.variables.state_variable import StateVariable
from slither.slithir.variables.reference import ReferenceVariable
from slither.slithir.variables.reference_ssa import ReferenceVariableSSA
from slither.slithir.variables.state_variable import StateIRVariable


class Delete(OperationWithLValue):
    """
    Delete has a lvalue, as it has for effect to change the value
    of its operand
    """

    def __init__(
        self,
        lvalue: StateIRVariable | StateVariable | ReferenceVariable,
        variable: StateIRVariable | StateVariable | ReferenceVariable | ReferenceVariableSSA,
    ) -> None:
        assert is_valid_lvalue(variable)
        super().__init__()
        self._variable = variable
        self._lvalue = lvalue

    @property
    def read(
        self,
    ) -> list[StateIRVariable | ReferenceVariable | ReferenceVariableSSA | StateVariable]:
        return [self.variable]

    @property
    def variable(
        self,
    ) -> StateIRVariable | StateVariable | ReferenceVariable | ReferenceVariableSSA:
        return self._variable

    def __str__(self) -> str:
        return f"{self.lvalue} = delete {self.variable} "



import abc
import re
from logging import Logger
from typing import TYPE_CHECKING
from collections.abc import Callable

from slither.core.compilation_unit import SlitherCompilationUnit, Language
from slither.core.declarations import Contract
from slither.formatters.exceptions import FormatImpossible
from slither.formatters.utils.patches import apply_patch, create_diff
from slither.utils.colors import green, yellow, red
from slither.utils.comparable_enum import ComparableEnum
from slither.utils.output import Output, SupportedOutput

if TYPE_CHECKING:
    from slither import Slither


class IncorrectDetectorInitialization(Exception):
    pass


class DetectorClassification(ComparableEnum):
    HIGH = 0
    MEDIUM = 1
    LOW = 2
    INFORMATIONAL = 3
    OPTIMIZATION = 4

    UNIMPLEMENTED = 999


classification_colors: dict[DetectorClassification, Callable[[str], str]] = {
    DetectorClassification.INFORMATIONAL: green,
    DetectorClassification.OPTIMIZATION: green,
    DetectorClassification.LOW: green,
    DetectorClassification.MEDIUM: yellow,
    DetectorClassification.HIGH: red,
}

classification_txt = {
    DetectorClassification.INFORMATIONAL: "Informational",
    DetectorClassification.OPTIMIZATION: "Optimization",
    DetectorClassification.LOW: "Low",
    DetectorClassification.MEDIUM: "Medium",
    DetectorClassification.HIGH: "High",
}


def make_solc_versions(minor: int, patch_min: int, patch_max: int) -> list[str]:
    """
    Create a list of solc version: [0.minor.patch_min .... 0.minor.patch_max]
    """
    return [f"0.{minor}.{x}" for x in range(patch_min, patch_max + 1)]


ALL_SOLC_VERSIONS_04 = make_solc_versions(4, 0, 26)
ALL_SOLC_VERSIONS_05 = make_solc_versions(5, 0, 17)
ALL_SOLC_VERSIONS_06 = make_solc_versions(6, 0, 12)
ALL_SOLC_VERSIONS_07 = make_solc_versions(7, 0, 6)
# No VERSIONS_08 as it is still in dev

DETECTOR_INFO = list[str | SupportedOutput]


class AbstractDetector(metaclass=abc.ABCMeta):
    ARGUMENT = ""  # run the detector with slither.py --ARGUMENT
    HELP = ""  # help information
    IMPACT: DetectorClassification = DetectorClassification.UNIMPLEMENTED
    CONFIDENCE: DetectorClassification = DetectorClassification.UNIMPLEMENTED

    WIKI = ""

    WIKI_TITLE = ""
    WIKI_DESCRIPTION = ""
    WIKI_EXPLOIT_SCENARIO = ""
    WIKI_RECOMMENDATION = ""

    STANDARD_JSON = True

    # list of vulnerable solc versions as strings (e.g. ["0.4.25", "0.5.0"])
    # If the detector is meant to run on all versions, use None
    VULNERABLE_SOLC_VERSIONS: list[str] | None = None
    # If the detector is meant to run on all languages, use None
    # Otherwise, use `solidity` or `vyper`
    LANGUAGE: str | None = None

    def __init__(
        self, compilation_unit: SlitherCompilationUnit, slither: "Slither", logger: Logger
    ) -> None:
        self.compilation_unit: SlitherCompilationUnit = compilation_unit
        self.contracts: list[Contract] = compilation_unit.contracts
        self.slither: Slither = slither
        # self.filename = slither.filename
        self.logger = logger

        if not self.HELP:
            raise IncorrectDetectorInitialization(
                f"HELP is not initialized {self.__class__.__name__}"
            )

        if not self.ARGUMENT:
            raise IncorrectDetectorInitialization(
                f"ARGUMENT is not initialized {self.__class__.__name__}"
            )

        if not self.WIKI:
            raise IncorrectDetectorInitialization(
                f"WIKI is not initialized {self.__class__.__name__}"
            )

        if not self.WIKI_TITLE:
            raise IncorrectDetectorInitialization(
                f"WIKI_TITLE is not initialized {self.__class__.__name__}"
            )

        if not self.WIKI_DESCRIPTION:
            raise IncorrectDetectorInitialization(
                f"WIKI_DESCRIPTION is not initialized {self.__class__.__name__}"
            )

        if not self.WIKI_EXPLOIT_SCENARIO and self.IMPACT not in [
            DetectorClassification.INFORMATIONAL,
            DetectorClassification.OPTIMIZATION,
        ]:
            raise IncorrectDetectorInitialization(
                f"WIKI_EXPLOIT_SCENARIO is not initialized {self.__class__.__name__}"
            )

        if not self.WIKI_RECOMMENDATION:
            raise IncorrectDetectorInitialization(
                f"WIKI_RECOMMENDATION is not initialized {self.__class__.__name__}"
            )

        if self.VULNERABLE_SOLC_VERSIONS is not None and not self.VULNERABLE_SOLC_VERSIONS:
            raise IncorrectDetectorInitialization(
                f"VULNERABLE_SOLC_VERSIONS should not be an empty list {self.__class__.__name__}"
            )

        if self.LANGUAGE is not None and self.LANGUAGE not in [
            Language.SOLIDITY.value,
            Language.VYPER.value,
        ]:
            raise IncorrectDetectorInitialization(
                f"LANGUAGE should not be either 'solidity' or 'vyper' {self.__class__.__name__}"
            )

        if re.match("^[a-zA-Z0-9_-]*$", self.ARGUMENT) is None:
            raise IncorrectDetectorInitialization(
                f"ARGUMENT has illegal character {self.__class__.__name__}"
            )

        if self.IMPACT not in [
            DetectorClassification.LOW,
            DetectorClassification.MEDIUM,
            DetectorClassification.HIGH,
            DetectorClassification.INFORMATIONAL,
            DetectorClassification.OPTIMIZATION,
        ]:
            raise IncorrectDetectorInitialization(
                f"IMPACT is not initialized {self.__class__.__name__}"
            )

        if self.CONFIDENCE not in [
            DetectorClassification.LOW,
            DetectorClassification.MEDIUM,
            DetectorClassification.HIGH,
            DetectorClassification.INFORMATIONAL,
            DetectorClassification.OPTIMIZATION,
        ]:
            raise IncorrectDetectorInitialization(
                f"CONFIDENCE is not initialized {self.__class__.__name__}"
            )

    def _log(self, info: str) -> None:
        if self.logger:
            self.logger.info(self.color(info))

    def _is_applicable_detector(self) -> bool:
        if self.VULNERABLE_SOLC_VERSIONS:
            return (
                self.compilation_unit.is_solidity
                and self.compilation_unit.solc_version in self.VULNERABLE_SOLC_VERSIONS
            )
        if self.LANGUAGE:
            return self.compilation_unit.language.value == self.LANGUAGE
        return True

    @abc.abstractmethod
    def _detect(self) -> list[Output]:
        """TODO Documentation"""
        return []

    def detect(self) -> list[dict]:
        results: list[dict] = []

        # check solc version
        if not self._is_applicable_detector():
            return results

        # only keep valid result, and remove duplicate
        # Keep only dictionaries
        for r in [output.data for output in self._detect()]:
            if self.compilation_unit.core.valid_result(r) and r not in results:
                results.append(r)
        if results and self.logger:
            self._log_result(results)
        if self.compilation_unit.core.generate_patches:
            for result in results:
                try:
                    self._format(self.compilation_unit, result)
                    if "patches" not in result:
                        continue
                    result["patches_diff"] = {}
                    for file in result["patches"]:
                        original_txt = self.compilation_unit.core.source_code[file].encode("utf8")
                        patched_txt = original_txt
                        offset = 0
                        patches = result["patches"][file]
                        patches.sort(key=lambda x: x["start"])
                        if not all(
                            patches[i]["end"] <= patches[i + 1]["end"]
                            for i in range(len(patches) - 1)
                        ):
                            self._log(
                                f"Impossible to generate patch; patches collisions: {patches}"
                            )
                            continue
                        for patch in patches:
                            patched_txt, offset = apply_patch(patched_txt, patch, offset)
                        diff = create_diff(self.compilation_unit, original_txt, patched_txt, file)
                        if not diff:
                            self._log(f"Impossible to generate patch; empty {result}")
                        else:
                            result["patches_diff"][file] = diff

                except FormatImpossible as exception:
                    self._log(f"\nImpossible to patch:\n\t{result['description']}\t{exception}")

        if results and self.slither.triage_mode:
            while True:
                indexes = input(
                    f'Results to hide during next runs: "0,1,...,{len(results)}" or "All" (enter to not hide results):\n'
                )
                if indexes == "All":
                    self.slither.save_results_to_hide(results)
                    return []
                if indexes == "":
                    return results
                if indexes.startswith("["):
                    indexes = indexes[1:]
                if indexes.endswith("]"):
                    indexes = indexes[:-1]
                try:
                    indexes_converted = [int(i) for i in indexes.split(",")]
                    self.slither.save_results_to_hide(
                        [r for (idx, r) in enumerate(results) if idx in indexes_converted]
                    )
                    return [r for (idx, r) in enumerate(results) if idx not in indexes_converted]
                except ValueError:
                    self.logger.error(yellow("Malformed input. Example of valid input: 0,1,2,3"))
        results = sorted(results, key=lambda x: x["id"])

        return results

    @property
    def color(self) -> Callable[[str], str]:
        return classification_colors[self.IMPACT]

    def generate_result(
        self,
        info: DETECTOR_INFO,
        additional_fields: dict | None = None,
    ) -> Output:
        output = Output(
            info,
            additional_fields,
            standard_format=self.STANDARD_JSON,
            markdown_root=self.slither.markdown_root,
        )

        output.data["check"] = self.ARGUMENT
        output.data["impact"] = classification_txt[self.IMPACT]
        output.data["confidence"] = classification_txt[self.CONFIDENCE]

        return output

    @staticmethod
    def _format(_compilation_unit: SlitherCompilationUnit, _result: dict) -> None:
        """Implement format"""
        return

    def _log_result(self, results: list[dict]) -> None:
        info = "\n"
        info += f"Detector: {self.ARGUMENT}\n"
        for idx, result in enumerate(results):
            if self.slither.triage_mode:
                info += f"{idx}: "
            info += result["description"]
        info += f"Reference: {self.WIKI}"
        self._log(info)



from slither.slithir.operations.lvalue import OperationWithLValue


class TmpNewStructure(OperationWithLValue):
    def __init__(self, contract_name, lvalue):
        super().__init__()
        self._contract_name = contract_name
        self._lvalue = lvalue

    @property
    def contract_name(self):
        return self._contract_name

    @property
    def read(self):
        return []

    def __str__(self):
        return f"{self.lvalue} = tmpnew {self.contract_name}"



import io
import logging
import sys


class CapturingStringIO(io.StringIO):
    """
    I/O implementation which captures output, and optionally mirrors it to the original I/O stream it replaces.
    """

    def __init__(self, original_io=None):
        super().__init__()
        self.original_io = original_io

    def write(self, s):
        super().write(s)
        if self.original_io:
            self.original_io.write(s)


class StandardOutputCapture:
    """
    Redirects and captures standard output/errors.
    """

    original_stdout = None
    original_stderr = None
    original_logger_handlers = None

    @staticmethod
    def enable(block_original: bool = True) -> None:
        """
        Redirects stdout and stderr to a capturable StringIO.
        :param block_original: If True, blocks all output to the original stream. If False, duplicates output.
        :return: None
        """
        # Redirect stdout
        if StandardOutputCapture.original_stdout is None:
            StandardOutputCapture.original_stdout = sys.stdout
            sys.stdout = CapturingStringIO(
                None if block_original else StandardOutputCapture.original_stdout
            )

        # Redirect stderr
        if StandardOutputCapture.original_stderr is None:
            StandardOutputCapture.original_stderr = sys.stderr
            sys.stderr = CapturingStringIO(
                None if block_original else StandardOutputCapture.original_stderr
            )

            # Backup and swap root logger handlers
            root_logger = logging.getLogger()
            StandardOutputCapture.original_logger_handlers = root_logger.handlers
            root_logger.handlers = [logging.StreamHandler(sys.stderr)]

    @staticmethod
    def disable() -> None:
        """
        Disables redirection of stdout/stderr, if previously enabled.
        :return: None
        """
        # If we have a stdout backup, restore it.
        if StandardOutputCapture.original_stdout is not None:
            sys.stdout.close()
            sys.stdout = StandardOutputCapture.original_stdout
            StandardOutputCapture.original_stdout = None

        # If we have an stderr backup, restore it.
        if StandardOutputCapture.original_stderr is not None:
            sys.stderr.close()
            sys.stderr = StandardOutputCapture.original_stderr
            StandardOutputCapture.original_stderr = None

        # Restore our logging handlers
        if StandardOutputCapture.original_logger_handlers is not None:
            root_logger = logging.getLogger()
            root_logger.handlers = StandardOutputCapture.original_logger_handlers
            StandardOutputCapture.original_logger_handlers = None

    @staticmethod
    def get_stdout_output() -> str:
        """
        Obtains the output from the currently set stdout
        :return: Returns stdout output as a string
        """
        sys.stdout.seek(0)
        return sys.stdout.read()

    @staticmethod
    def get_stderr_output() -> str:
        """
        Obtains the output from the currently set stderr
        :return: Returns stderr output as a string
        """
        sys.stderr.seek(0)
        return sys.stderr.read()



from slither.core.variables.local_variable import LocalVariable
from slither.core.variables.state_variable import StateVariable

from slither.core.declarations.solidity_variables import SolidityVariable
from slither.core.variables.top_level_variable import TopLevelVariable

from slither.slithir.variables.temporary import TemporaryVariable
from slither.slithir.variables.constant import Constant
from slither.slithir.variables.reference import ReferenceVariable
from slither.slithir.variables.tuple import TupleVariable
from slither.core.source_mapping.source_mapping import SourceMapping

RVALUE = (
    StateVariable
    | LocalVariable
    | TopLevelVariable
    | TemporaryVariable
    | Constant
    | SolidityVariable
    | ReferenceVariable
)

LVALUE = StateVariable | LocalVariable | TemporaryVariable | ReferenceVariable | TupleVariable


def is_valid_rvalue(v: SourceMapping | None) -> bool:
    return isinstance(
        v,
        (
            StateVariable,
            LocalVariable,
            TopLevelVariable,
            TemporaryVariable,
            Constant,
            SolidityVariable,
            ReferenceVariable,
        ),
    )


def is_valid_lvalue(v: SourceMapping | None) -> bool:
    return isinstance(
        v,
        (
            StateVariable,
            LocalVariable,
            TemporaryVariable,
            ReferenceVariable,
            TupleVariable,
        ),
    )



from functools import partial
import platform
import sys


class Colors:
    COLORIZATION_ENABLED = True
    RED = "\033[91m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    BLUE = "\033[94m"
    MAGENTA = "\033[95m"
    BOLD = "\033[1m"
    END = "\033[0m"


def colorize(color: Colors, txt: str) -> str:
    if Colors.COLORIZATION_ENABLED:
        return f"{color}{txt}{Colors.END}"
    return txt


def enable_windows_virtual_terminal_sequences() -> bool:
    """
    Sets the appropriate flags to enable virtual terminal sequences in a Windows command prompt.
    Reference: https://docs.microsoft.com/en-us/windows/console/console-virtual-terminal-sequences
    """

    try:
        from ctypes import windll, byref  # type: ignore
        from ctypes.wintypes import DWORD, HANDLE

        kernel32 = windll.kernel32
        virtual_terminal_flag = 0x04  # ENABLE_VIRTUAL_TERMINAL_PROCESSING

        # Obtain our stdout/stderr handles.
        # Reference: https://docs.microsoft.com/en-us/windows/console/getstdhandle
        handle_stdout = kernel32.GetStdHandle(-11)
        handle_stderr = kernel32.GetStdHandle(-12)

        # Loop for each stdout/stderr handle.
        for current_handle in [handle_stdout, handle_stderr]:
            # If we get a null handle, or fail any subsequent calls in this scope, we do not colorize any output.
            if current_handle is None or current_handle == HANDLE(-1):
                return False

            # Try to obtain the current flags for the console.
            current_mode = DWORD()
            if not kernel32.GetConsoleMode(current_handle, byref(current_mode)):
                return False

            # If the virtual terminal sequence processing is not yet enabled, we enable it.
            if (current_mode.value & virtual_terminal_flag) == 0:
                if not kernel32.SetConsoleMode(
                    current_handle, current_mode.value | virtual_terminal_flag
                ):
                    return False
    except Exception:
        # Any generic failure (possibly from calling these methods on older Windows builds where they do not exist)
        # will fall back onto disabling colorization.
        return False

    return True


def set_colorization_enabled(enabled: bool) -> None:
    """
    Sets the enabled state of output colorization.
    :param enabled: Boolean indicating whether output should be colorized.
    :return: None
    """
    # If color is supposed to be enabled and this is windows, we have to enable console virtual terminal sequences:
    if enabled and platform.system() == "Windows":
        Colors.COLORIZATION_ENABLED = enable_windows_virtual_terminal_sequences()
    else:
        # This is not windows, or colorization is being disabled, so we can adjust the state immediately.
        Colors.COLORIZATION_ENABLED = enabled


green = partial(colorize, Colors.GREEN)
yellow = partial(colorize, Colors.YELLOW)
red = partial(colorize, Colors.RED)
blue = partial(colorize, Colors.BLUE)
magenta = partial(colorize, Colors.MAGENTA)
bold = partial(colorize, Colors.BOLD)

# We enable colorization by default if the output is a tty
set_colorization_enabled(sys.stdout.isatty())



"""
Module detecting unimplemented functions
Recursively check the called functions

Collect all the implemented and unimplemented functions of all the contracts
Check for unimplemented functions that are never implemented
Consider public state variables as implemented functions
Do not consider fallback function or constructor
"""

from slither.core.declarations import Function
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.core.declarations.contract import Contract
from slither.core.declarations.function_contract import FunctionContract
from slither.utils.output import Output


# Since 0.5.1, Solidity allows creating state variable matching a function signature.
older_solc_versions = ["0.5.0"] + ["0.4." + str(x) for x in range(0, 27)]


class UnimplementedFunctionDetection(AbstractDetector):
    """
    Unimplemented functions detector
    """

    ARGUMENT = "unimplemented-functions"
    HELP = "Unimplemented functions"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#unimplemented-functions"

    WIKI_TITLE = "Unimplemented functions"
    WIKI_DESCRIPTION = "Detect functions that are not implemented on derived-most contracts."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
interface BaseInterface {
    function f1() external returns(uint);
    function f2() external returns(uint);
}

interface BaseInterface2 {
    function f3() external returns(uint);
}

contract DerivedContract is BaseInterface, BaseInterface2 {
    function f1() external returns(uint){
        return 42;
    }
}
```
`DerivedContract` does not implement `BaseInterface.f2` or `BaseInterface2.f3`.
As a result, the contract will not properly compile.
All unimplemented functions must be implemented on a contract that is meant to be used."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Implement all unimplemented functions in any contract you intend to use directly (not simply inherit from)."

    @staticmethod
    def _match_state_variable(contract: Contract, f: FunctionContract) -> bool:
        return any(s.full_name == f.full_name for s in contract.state_variables)

    def _detect_unimplemented_function(self, contract: Contract) -> set[Function]:
        """
        Detects any function definitions which are not implemented in the given contract.
        :param contract: The contract to search unimplemented functions for.
        :return: A list of functions which are not implemented.
        """

        # If it's simply a contract signature, we have no functions.
        if contract.is_signature_only():
            return set()

        # Populate our unimplemented functions set with any functions not implemented in this contract, excluding the
        # fallback function and constructor.
        unimplemented = set()
        for f in contract.all_functions_called:
            if not isinstance(f, Function):
                continue
            if (
                not f.is_implemented
                and not f.is_constructor
                and not f.is_fallback
                and not f.is_constructor_variables
            ):
                if self.compilation_unit.solc_version not in older_solc_versions:
                    # Since 0.5.1, Solidity allows creating state variable matching a function signature
                    if not self._match_state_variable(contract, f):
                        unimplemented.add(f)
                else:
                    unimplemented.add(f)
        return unimplemented

    def _detect(self) -> list[Output]:
        """Detect unimplemented functions

        Recursively visit the calls
        Returns:
            list: {'vuln', 'filename,'contract','func'}
        """
        results = []
        for contract in self.compilation_unit.contracts_derived:
            functions = self._detect_unimplemented_function(contract)
            if functions:
                info: DETECTOR_INFO = [contract, " does not implement functions:\n"]

                for function in sorted(functions, key=lambda x: x.full_name):
                    info += ["\t- ", function, "\n"]

                res = self.generate_result(info)
                results.append(res)
        return results



"""
Module detecting assignment of array length
"""

from slither.core.variables import Variable
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    ALL_SOLC_VERSIONS_04,
    ALL_SOLC_VERSIONS_05,
)
from slither.core.cfg.node import Node, NodeType
from slither.slithir.operations import Assignment, Length
from slither.slithir.variables.reference import ReferenceVariable
from slither.slithir.operations.binary import Binary
from slither.analyses.data_dependency.data_dependency import is_tainted
from slither.core.declarations.contract import Contract
from slither.utils.output import Output, SupportedOutput


def detect_array_length_assignment(contract: Contract) -> set[Node]:
    """
    Detects and returns all nodes which assign array length.
    :param contract: Contract to detect assignment within.
    :return: A list of tuples with (Variable, node) where Variable references an array whose length was set by node.
    """

    # Create our result set.
    results = set()

    # Loop for each function and modifier.

    for function in contract.functions_and_modifiers_declared:
        # Define a set of reference variables which refer to array length.
        array_length_refs = set()

        # Loop for every node in this function, looking for expressions where array length references are made,
        # and subsequent expressions where array length references are assigned to.
        for node in function.nodes:
            if node.type == NodeType.EXPRESSION:
                for ir in node.irs:
                    # First we look for the member access for 'length', for which a reference is created.
                    # We add the reference to our list of array length references.
                    if isinstance(ir, Length):  # a
                        #                            if ir.variable_right == "length":
                        array_length_refs.add(ir.lvalue)

                    # If we have an assignment/binary operation, verify the left side refers to a reference variable
                    # which is in our list or array length references. (Array length is being assigned to).
                    elif isinstance(ir, (Assignment, Binary)):
                        if isinstance(ir.lvalue, ReferenceVariable):
                            if ir.lvalue in array_length_refs and any(
                                is_tainted(v, contract) for v in ir.read if isinstance(v, Variable)
                            ):
                                # the taint is not precise enough yet
                                # as a result, REF_0 = REF_0 + 1
                                # where REF_0 points to a LENGTH operation
                                # is considered as tainted
                                if ir.lvalue in ir.read:
                                    continue
                                results.add(node)
                                break

    # Return the resulting set of nodes which set array length.
    return results


class ArrayLengthAssignment(AbstractDetector):
    """
    Array length assignment
    """

    ARGUMENT = "controlled-array-length"
    HELP = "Tainted array length assignment"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#array-length-assignment"

    WIKI_TITLE = "Array Length Assignment"
    WIKI_DESCRIPTION = """Detects the direct assignment of an array's length."""

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract A {
	uint[] testArray; // dynamic size array

	function f(uint usersCount) public {
		// ...
		testArray.length = usersCount;
		// ...
	}

	function g(uint userIndex, uint val) public {
		// ...
		testArray[userIndex] = val;
		// ...
	}
}
```
Contract storage/state-variables are indexed by a 256-bit integer.
The user can set the array length to `2**256-1` in order to index all storage slots.
In the example above, one could call the function `f` to set the array length, then call the function `g` to control any storage slot desired.
Note that storage slots here are indexed via a hash of the indexers; nonetheless, all storage will still be accessible and could be controlled by the attacker."""
    # endregion wiki_exploit_scenario

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """Do not allow array lengths to be set directly set; instead, opt to add values as needed.
Otherwise, thoroughly review the contract to ensure a user-controlled variable cannot reach an array length assignment."""
    # endregion wiki_recommendation

    VULNERABLE_SOLC_VERSIONS = ALL_SOLC_VERSIONS_04 + ALL_SOLC_VERSIONS_05

    def _detect(self) -> list[Output]:
        """
        Detect array length assignments
        """
        results = []
        for contract in self.contracts:
            array_length_assignments = detect_array_length_assignment(contract)
            if array_length_assignments:
                contract_info: list[str | SupportedOutput] = [
                    contract,
                    " contract sets array length with a user-controlled value:\n",
                ]
                for node in array_length_assignments:
                    node_info: list[str | SupportedOutput] = contract_info + [
                        "\t- ",
                        node,
                        "\n",
                    ]
                    res = self.generate_result(node_info)
                    results.append(res)

        return results
